/*
Daybo Logic Run-Time TarBall Library And Tools
Copyright (c) 2000-2006, David Duncan Ross Palmer, Daybo Logic
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
      
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
      
    * Neither the name of the Daybo Logic nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

/*
  RTTB library core, protected data should go here and non-user callable
  functions for the rest of the library's use.  You'll find most internal
  workings here, it's the bulk of the library.
*/

#define RTTB_SOURCE
#define RTTB_CORE
#define RTTB_USING_TYPES

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif /*HAVE_CONFIG_H*/

#include <time.h>
#include <stdio.h>
#include <assert.h>

#ifdef HAVE_STDLIB_H
# include <stdlib.h>
#endif /*HAVE_STDLIB_H*/

#ifdef HAVE_STRING_H
# include <string.h>
#endif /*HAVE_STRING_H*/

#ifdef HAVE_LIMITS_H
# include <limits.h>
#endif /*HAVE_LIMITS_H*/

#ifdef HAVE_STDBOOL_H
# include <stdbool.h>
#else
# include "dlstdinc_stdbool.h"
#endif /*HAVE_STDBOOL_H*/

#ifdef RTTB_HDRSTOP
# pragma hdrstop
#endif /*RTTB_HDRSTOP*/

#include "rttb.h" /* Public header */
#include "rttb_int.h" /* Internal header */
#include "rttb_ll.h" /* Linked list (generic) */
#include "rttb_strzz.h" /* String list */
/*-------------------------------------------------------------------------*/
/*
  12th April 2001 use this instead of -1L to check if long int values are
  really invalid and need reloading rather than checking the value directly,
  incase the value changes
*/
#define CACHE_INVALID (-1L)

/* Private core types */
/* Cached header information */
struct header {
  BYTE revision; /* Revision number of RTTB format */
  WORD32 n; /* Number of files in the archive */
  WORD32 fastDataMarker; /* Internal use only, see documentation */
  struct link_ptrs lengths; /* Lengths of files (linked list of long ints) */
  struct link_ptrs checksums; /* Linked list of 32-bit checksums for all files */
  strzz_t fileNames; /* Filenames (Win32 style double NULL terminated string) */
};

struct rttbfile {
  RTTB_HANDLE rttbHandle; /* Invalid entry if RTTB_NULLHANDLE, ignore */
  BOUNDFILE boundfile;  /* These caches added 4th Apr 2001 - do not read directly */
  long int lenBytes; /* Length of file or fake contained file */
  long int activeMarker; /* Relative to container or, if loose, absolute cached ftell() */
};

struct rttbcontainer {
  BOUNDFILE boundfile;
  struct rttbfile openFileArray[RTTB_MAXOPEN_PERRTTB];
  struct header cachedHeader;
  int prepared;
  /* --New caching fields as of 4th April 2001, do not read directly-- */
  long int lenBytes; /* Length of the entire container including header (cached) */
};

/*
  These used to be called directly from outside,
  now they cannot be but through rttb_core()
*/
static void rttb_core_Init(void);

static void rttb_core_Uninit(void);

static inline int rttb_core_IsStarted(void);

static void rttb_core_InitBoundFile(
  BOUNDFILE* p
);

static RTTB_HANDLE rttb_core_GenerateHandle(void);

static unsigned int rttb_core_AddHandle(
  const RTTB_HANDLE Handle,
  const char* ContainerFileName,
  const BOUNDFILE* BoundFile
);

static int rttb_core_FreeHandle(
  const RTTB_HANDLE Handle
);

static FILE* rttb_core_GetFSP(
  const RTTB_HANDLE Handle
);

static int rttb_core_RegisterContainer(
  const char* ContainerFileName
);

static int rttb_core_UnregisterContainer(
  const char* ContainerFileName
);

static RTTB_ERROR rttb_core_PrepareContainer(
  const char* ContainerFileName
);

static int rttb_core_IsContainerRegistered(
  const char* ContainerFileName
);

static int rttb_core_IsContainerPrepared(
  const char* ContainerFileName
);

static int rttb_core_UnprepareContainer(
  const char* ContainerFileName
);

static RTTB_ERROR rttb_core_PrepareFile(
  const char* RTTB, const char* FileName
);

static int rttb_core_CheckFileExists(
  const char* RTTB,
  const char* FileName
);

static int rttb_core_IsValidHandle(
  const RTTB_HANDLE Handle
);

static void rttb_core_SetFileMode(
  const RTTB_HANDLE RTTBHandle,
  const char* Mode
);

static const char* rttb_core_GetFileMode(
  const RTTB_HANDLE RTTBHandle
);

static RTTB_HANDLE rttb_core_GetHandle(
  const char* RTTB,
  const char* FileName
);

static RTTB_ERROR rttb_core_GetFileOffset(
  const RTTB_HANDLE Handle,
  long int* POffset
);

static RTTB_ERROR rttb_core_SetFileOffset(
  const RTTB_HANDLE Handle,
  const long int NewOffsetRecord
);

static RTTB_ERROR rttb_core_GetFileLength(
  const RTTB_HANDLE Handle,
  long int* PLength
);

static size_t rttb_core_ContainedRead(
  RTTB_HANDLE Handle,
  size_t Size,
  size_t N,
  void* PTarget,
  RTTB_ERROR* PError
);

static long int rttb_core_LocalToGlobalOffsetByHandle(
  RTTB_HANDLE Handle,
  long int LocalOffset
);

static RTTB_BYTE rttb_core_ReadOneByte(
  RTTB_HANDLE Handle,
  RTTB_ERROR* PError
);

static char* rttb_core_ReadString(
  RTTB_HANDLE Handle,
  char* s,
  unsigned int size,
  RTTB_ERROR* PError
);

/* Actual storage of entrypoint table */
static const struct _rttb_core_entries _entrypoints = {
  rttb_core_Init,
  rttb_core_Uninit,
  rttb_core_IsStarted,
  rttb_core_InitBoundFile,
  rttb_core_GenerateHandle,
  rttb_core_AddHandle,
  rttb_core_FreeHandle,
  rttb_core_GetFSP,
  rttb_core_RegisterContainer,
  rttb_core_UnregisterContainer,
  rttb_core_PrepareContainer,
  rttb_core_IsContainerRegistered,
  rttb_core_IsContainerPrepared,
  rttb_core_UnprepareContainer,
  rttb_core_PrepareFile,
  rttb_core_CheckFileExists,
  rttb_core_IsValidHandle,
  rttb_core_SetFileMode,
  rttb_core_GetFileMode,
  rttb_core_GetHandle,
  rttb_core_GetFileOffset,
  rttb_core_SetFileOffset,
  rttb_core_GetFileLength,
  rttb_core_ContainedRead,
  rttb_core_LocalToGlobalOffsetByHandle,
  rttb_core_ReadOneByte,
  rttb_core_ReadString
}; /*_entrypoints*/

static int _started = 0; /* Flag indicating whether library has been initalized */
static struct link_ptrs _rttblist; /* Linked list of rttbcontainers */

static int _i_checkhandle(
  const RTTB_HANDLE Handle
);

static inline int _i_checkfsp(
  FILE* fsp
);

static int _i_checkrttbname(
  const char* Filename,
  unsigned int* PIndex
);

static RTTB_ERROR _i_readrttbheader(
  struct rttbcontainer* Container
);

/*
  Returns a pointer to the first space for a file entry which is not used
  or NULL if they are all in use
*/
static struct rttbfile* _i_freefile(
  struct rttbcontainer* Container
);

static unsigned int _i_assosiationcount(
  struct rttbcontainer* Container
);

/*
  Returns the start offset of a file within the container
   - JUST DOES CALCULATIONS, NEEDS TO BE SPED UP WITH CACHING
*/
static long int _i_getphysicalpos(
  struct rttbcontainer* Container,
  const RTTB_UINT MasterListIndex,
  const RTTB_UINT Index,
  bool *PModFlag
);

/*
  Reads data from any file, it doesn't update offset caches so watch out
*/
static RTTB_ERROR _i_directread(
  FILE* FileStream,
  size_t Bytes,
  void* RawData
);

/*
  Returns the fast data marker for a specified container
*/
static inline long int _i_getfastdata(
  struct rttbcontainer* Container
);

static long int _i_LocalToGlobalOffset(
  struct rttbcontainer* Container,
  RTTB_UINT Index,
  long int LocalOffset
);

/*
  When a handle is found and the actual index is a container file is needed,
  look it up with this helper
*/
static RTTB_UINT _i_OpenArrayToHeaderIndex(
  struct rttbcontainer* Container,
  RTTB_UINT Index
);

static RTTB_UINT _i_HeaderIndexToOpenArray(
  struct rttbcontainer* Container,
  RTTB_UINT HeaderIndex
);

/*
  Virtually get a file offset (activeMarker) from the index into the
  cached header's linked list(s)
*/
static long int _i_GetFileOffInCacheHead(
  struct rttbcontainer* Container,
  RTTB_UINT HeaderIndex
);

/*
  Get a filename from a cached header
*/
static inline const char* _i_GetFNPtrInCacheHead(
  struct rttbcontainer* Container,
  RTTB_UINT HeaderIndex
);

static RTTB_BYTE _i_readbyte(
  FILE* FileStream,
  RTTB_ERROR* PError
);

/*
  Seeks to a position in a container, if a container is held on the stack
  the caller should write the container back to the _rttblist, failure
  to do so will result in corruption of the container offset cache
  which will cause strange trouble later.  If the seek function
  returns an error, the seek will have been aborted and the file
  offset will be the same as before, therefore there is no need to
  write back to _rttblist.
*/
static RTTB_ERROR _i_container_seek(
  struct rttbcontainer* Container,
  long int Offset
);

static long int _i_loose_getfileoffset(
  struct rttbcontainer* PContainer,
  const unsigned int Index,
  bool *PModFlag
);

static long int _i_contained_getfileoffset(
  struct rttbcontainer* PContainer,
  const unsigned int Index,
  bool *PModFlag
);

static long int _i_container_getfileoffset(
  struct rttbcontainer* PContainer,
  bool *PModFlag
);

static void _i_loose_setfileoffset(
  struct rttbcontainer* PContainer,
  const unsigned int Index,
  const long int Offset
);

/* Same code in both!  Merged to save overhead : 12th Apr 2001 */
#define _i_contained_setfileoffset _i_loose_setfileoffset

static long int _i_loose_getfilelength(
  struct rttbcontainer* PContainer,
  const unsigned int Index,
  bool *PModFlag
);

static long int _i_contained_getfilelength(
  struct rttbcontainer* PContainer,
  const unsigned int Index,
  bool *PModFlag
);

static long int _i_container_getfilelength(
  struct rttbcontainer* PContainer,
  bool *PModFlag
);

/* Note: No file length setters because this is a readonly library */

/* Structure initializers */
static void _i_initheader(struct header* p);
static void _i_initrttbfile(struct rttbfile* p);
static void _i_initrttbcontainer(struct rttbcontainer* p);
/* Note rttb_core_InitBoundFile() is in rttb_int.h */
/* Stuff for when finished with structures, they auto-re-init too */
static void _i_freeheader(struct header* p);
static void _i_freerttbfile(struct rttbfile* p);
static void _i_freeboundfile(BOUNDFILE* p);
static void _i_freerttbcontainer(struct rttbcontainer* p);
static void _i_releaseallhandles(struct rttbcontainer* p);
static void _i_releaseallcontainers(void);

/*
  Misc helpers - these don't touch data directly and could be moved out
  of the core.
*/
static char* _i_setstr(
  char* str,
  const char* NewString
);

/*
  Position recovery after bad seek, loose files only
*/
static inline void _i_recover(
  FILE* fsp,
  const long int OldPos
);

/*
  Gets the physical file length of a loose (blah) file,
  could be used on a container if neccersary.
  FIXME: Uses a bad technique, use fstat() instead,
  according to the chaps in #C on irc://irc.freenode.net
*/
static long int _i_getphysicallength(
  FILE* fsp
);

/*--------------------------------------------------------------------------*/
inline const struct _rttb_core_entries* rttb_core()
{
  return &_entrypoints;
}
/*--------------------------------------------------------------------------*/
static void rttb_core_Init()
{
  assert(!_started);
  rttb_link_ptrs_init(&_rttblist, sizeof(struct rttbcontainer));
  _started = 1;
  return;
}
/*--------------------------------------------------------------------------*/
static void rttb_core_Uninit()
{
  assert(_started);
  _started = 0;
  _i_releaseallcontainers();
}
/*--------------------------------------------------------------------------*/
inline static int rttb_core_IsStarted()
{
  return _started;
}
/*--------------------------------------------------------------------------*/
static RTTB_HANDLE rttb_core_GenerateHandle()
{
  /* Return the first available handle */
  RTTB_HANDLE testHandle = RTTB_NULLHANDLE;

  while (1) {
    testHandle++; /* Try next valud handle (or make not RTTB_NULLHANDLE) */
    if ( !_i_checkhandle(testHandle) ) /* Not used */
      break; /* They can have this */
    if ( testHandle == RTTB_MAXHANDLE )
      return RTTB_NULLHANDLE; /* All used, they are out of luck */
  }
  return testHandle;
}
/*-------------------------------------------------------------------------*/
static unsigned int rttb_core_AddHandle(
  const RTTB_HANDLE Handle,
  const char* ContainerFileName,
  const BOUNDFILE* BoundFile
)
{
  struct rttbcontainer container;
  unsigned int containerIndex, i;
  assert(_started);
  if ( Handle == RTTB_NULLHANDLE )
    return 3; /* Attempt to add RTTB_NULLHANDLE, illegal */
  if ( _i_checkhandle(Handle) )
    return 1; /* The handle is not unique */  
  if ( !_i_checkfsp(BoundFile->pfs) )
    return 2; /* File stream pointer not valid */
  if ( !_i_checkrttbname(ContainerFileName, &containerIndex) )
    return 4; /* The specified RTTB container is not valid */
  _rttblist.Get(&_rttblist, containerIndex, &container); /* Read back container */
  /* Find first free handle */
  for ( i = 0U; i < RTTB_MAXOPEN_PERRTTB; i++ ) {
    if ( !container.openFileArray[i].rttbHandle ) { /* First free entry */
      _i_initrttbfile(&container.openFileArray[i]);
      container.openFileArray[i].rttbHandle = Handle;
      container.openFileArray[i].boundfile = *BoundFile;
      /* Write back the container */
      _rttblist.Set(&_rttblist, containerIndex, &container);
      return 1; /* Success */
    }
  }
  return 0; /* Failed */
}
/*-------------------------------------------------------------------------*/
static int rttb_core_FreeHandle(
  const RTTB_HANDLE Handle
)
{
  unsigned int i, c;
  assert(_started);
  assert( Handle != RTTB_NULLHANDLE ); /* Make sure I'm not accidently trying to free blank slots in the container */
  c = _rttblist.Count(&_rttblist);

  for ( i = 0U; i < c; i++ ) { /* All containers loop */
    struct rttbcontainer container;
    unsigned int j;
    bool modContainer = false; /* If it's set the container needs written back */
    _rttblist.Get(&_rttblist, i, &container);
    for ( j = 0U; j < RTTB_MAXOPEN_PERRTTB; j++ ) { /* All slots per container loop */
      if ( container.openFileArray[j].rttbHandle == Handle ) {
        _i_freerttbfile(&container.openFileArray[j]);
        modContainer = true;
        break; /* No duplicates allowed so I can quit here */
      }
    }
    if ( !_i_assosiationcount(&container) ) { /* No assosiated files anymore? */
      _i_freerttbcontainer(&container);
      modContainer = true;
    }
    if ( modContainer ) /* Container information has been modified */
      _rttblist.Set(&_rttblist, i, &container);
    return 1;
  }
  return 0;
}
/*-------------------------------------------------------------------------*/
static FILE* rttb_core_GetFSP(
  const RTTB_HANDLE Handle
)
{
  unsigned int i, c;

  assert(_started);
  c = _rttblist.Count(&_rttblist);
  for ( i = 0U; i < c; i++ ) { /* Containers */
    unsigned int j;
    struct rttbcontainer container;
    _rttblist.Get(&_rttblist, i, &container);
    for ( j = 0U; j < RTTB_MAXOPEN_PERRTTB; j++ ) {
      if ( container.openFileArray[j].rttbHandle == Handle )
        return container.openFileArray[j].boundfile.pfs;
    }
  }
  assert(0); /* Handle not valid, debug this */
  return NULL;
}
/*-------------------------------------------------------------------------*/
static int rttb_core_RegisterContainer(
  const char* ContainerFileName
)
{
  struct rttbcontainer newcontainer;

  assert(ContainerFileName);
  assert(_started);
#ifndef NDEBUG /* Debug code, we trust core users when in release mode */
  if (rttb_core_IsContainerRegistered(ContainerFileName)) return 0;
#endif /*!NDEBUG*/
  _i_initrttbcontainer(&newcontainer);
  newcontainer.boundfile.filename = (char*)rttb_malloc(strlen(ContainerFileName)+1);
  if ( !newcontainer.boundfile.filename ) return 0;
  strcpy(newcontainer.boundfile.filename, ContainerFileName);

  if (_rttblist.Add(&_rttblist, &newcontainer))
    return 1;

  /* Failed, clean up locally */
  _i_freerttbcontainer(&newcontainer);
  return 0;
}
/*-------------------------------------------------------------------------*/
static int rttb_core_UnregisterContainer(
  const char* ContainerFileName
)
{
  unsigned int i, c;

  assert(_started);
  assert(ContainerFileName);

  if ( !rttb_core_IsContainerRegistered(ContainerFileName) )
    return 0;
  if ( !rttb_core_IsContainerPrepared(ContainerFileName) )
    rttb_core_UnprepareContainer(ContainerFileName);

  c = _rttblist.Count(&_rttblist);
  for ( i = 0U; i < c; i++ ) {
    struct rttbcontainer container;

    _rttblist.Get(&_rttblist, i, &container);
    if ( container.boundfile.filename ) {
      if ( strcmp(container.boundfile.filename, ContainerFileName) == 0 ) {
        _i_freerttbcontainer(&container);
        _rttblist.DelByPos(&_rttblist, i);
        return 1;
      }
    }
  }
  return 0;
}
/*-------------------------------------------------------------------------*/
static RTTB_ERROR rttb_core_PrepareContainer(
  const char* ContainerFileName
)
{
  FILE* Hcont;
  unsigned int i, c;
  assert(_started);
  assert(ContainerFileName);
  c = _rttblist.Count(&_rttblist);
  for ( i = 0U; i < c; i++ ) {
    struct rttbcontainer container;
    _rttblist.Get(&_rttblist, i, &container);
    if ( container.boundfile.filename ) {
      if ( strcmp(container.boundfile.filename, ContainerFileName) == 0 ) {
        /* This is the one */
        if ( !container.prepared ) {
          /* Prepare the container by caching header etc. */
          Hcont = rttb_open(ContainerFileName, "rb");
          if (!Hcont) return RTTB_ERROR_CANTOPEN;
          container.boundfile.pfs = Hcont;
          if ( _i_readrttbheader(&container) != RTTB_ERROR_NOERROR ) { /* Could not read the header */
            _i_freerttbcontainer(&container);
            #ifndef NDEBUG /* Dump chunk info from memory in debug builds */
            rttb_debug_chunkdump();
            #endif /*!NDEBUG*/
            return RTTB_ERROR_UNEXPECTEDEOF;
          }
          container.prepared = 1; /* TRUE */
          _rttblist.Set(&_rttblist, i, &container); /* Write back changes */
          return RTTB_ERROR_NOERROR;
        }
      }
    }
  }
  /* If it gets here then the container wasn't even registered and cannot be prepared */
  return RTTB_ERROR_INVRTTB; /* Probally the best error for this */
}
/*-------------------------------------------------------------------------*/
static int rttb_core_IsContainerRegistered(
  const char* ContainerFileName
)
{
  unsigned int i, c;

  assert(_started);
  c = _rttblist.Count(&_rttblist);
  if ( ContainerFileName ) {
    for ( i = 0U; i < c; i++ ) {
      struct rttbcontainer container;

      _rttblist.Get(&_rttblist, i, &container);
      if ( container.boundfile.filename ) {
        if ( strcmp(container.boundfile.filename, ContainerFileName) == 0 )
          return 1; /* Yes, 'tis registered captain */
      }
    }
  }
  return 0; /* Not registered */
}
/*-------------------------------------------------------------------------*/
static int rttb_core_IsContainerPrepared(
  const char* ContainerFileName
)
{
  unsigned int i, c;

  c = _rttblist.Count(&_rttblist);
  for ( i = 0U; i < c; i++ )
  {
    struct rttbcontainer container;

    _rttblist.Get(&_rttblist, i, &container);
    if ( container.boundfile.filename )
    {
      if ( strcmp(container.boundfile.filename, ContainerFileName) == 0 )
        return container.prepared;
    }
  }
  return 0; /* Not found/not prepared */
}
/*-------------------------------------------------------------------------*/
static int rttb_core_UnprepareContainer(
  const char* ContainerFileName
)
{
  unsigned int i, c;

  c = _rttblist.Count(&_rttblist);
  for ( i = 0U; i < c; i++ )
  {
    struct rttbcontainer container;

    _rttblist.Get(&_rttblist, i, &container);
    if ( container.boundfile.filename )
    {
      if ( strcmp(container.boundfile.filename, ContainerFileName) == 0 )
      {
        _i_releaseallhandles(&container);
        assert(0); /* Debug, add more code here */
        _rttblist.Set(&_rttblist, i, &container);
        return 1;
      }
    }
  }
  return 0;
}
/*-------------------------------------------------------------------------*/
static int _i_checkhandle(
  const RTTB_HANDLE Handle
)
{
  unsigned int i, c;  c = _rttblist.Count(&_rttblist);
  for ( i = 0U; i < c; i++ ) {
    struct rttbcontainer container;
    unsigned int j;
    _rttblist.Get(&_rttblist, i, &container);
    for ( j = 0U; j < RTTB_MAXOPEN_PERRTTB; j++ ) {
      if ( container.openFileArray[j].rttbHandle == Handle )
        return 1; /* Handle present & correct */
    }
  }
  return 0;
}
/*-------------------------------------------------------------------------*/
static inline int _i_checkfsp(
  FILE* fsp
)
{
  if ( !fsp ) return 0;
  if ( ftell(fsp) == -1L ) return 0;
  return 1;
}
/*-------------------------------------------------------------------------*/
static RTTB_ERROR _i_readrttbheader(
  struct rttbcontainer* Container
)
{
  long int stackedFilePos;
  unsigned int i; /* General purpose, used a few times */
  /*WORD16 w16;*/ /* General purpose temporaries for reading binary values */
  WORD32 w32;
  WORD16* lenfns; /* Filename lengths dynamic array */
  long int wlong;
  unsigned int c; /* Count for loop speedups (only effective on non-optimizing compilers) */
  RTTB_ERROR status = RTTB_ERROR_NOERROR; /* Return value */

  assert(Container);
  if ( !_i_checkfsp(Container->boundfile.pfs) ) return 0; /* Bad FSP */

  _i_initheader(&Container->cachedHeader);
  stackedFilePos = ftell(Container->boundfile.pfs);
  rewind(Container->boundfile.pfs);  /* Read stuff from file NOW! */
  fseek(Container->boundfile.pfs, RTTB_COPYRIGHTLENGTH, SEEK_SET); /* Seek past egotistical banner */
  if ( fread(&Container->cachedHeader.revision, sizeof(Container->cachedHeader.revision), 1, Container->boundfile.pfs) < 1 ) {
    status = RTTB_ERROR_UNEXPECTEDEOF;
    goto readfail;
  }
#if ( RTTB_REVISION > 0 )
  if ( Container->header.revision < RTTB_REVISION ) { /* Old version, we can't support */
    status = RTTB_ERROR_OLD;
    goto readfail;
  }

  if ( Container->header.revision > RTTB_REVISION ) { /* New version we can't support either */
    status = RTTB_ERROR_NEW;
    goto readfail;
  }
#else
  if ( Container->cachedHeader.revision > RTTB_REVISION ) {
    status = RTTB_ERROR_NEW;
    goto readfail;
  }
#endif
  if ( fread(&Container->cachedHeader.n, sizeof(Container->cachedHeader.n), 1, Container->boundfile.pfs) < 1 ) {
    status = RTTB_ERROR_UNEXPECTEDEOF;
    goto readfail;
  }
#ifdef __L_ENDIAN__
  Container->cachedHeader.n = rttb_Swap32(Container->cachedHeader.n);
#endif /*__L_ENDIAN__*/
  if ( fread(&Container->cachedHeader.fastDataMarker, sizeof(Container->cachedHeader.fastDataMarker), 1, Container->boundfile.pfs) < 1 ) {
    status = RTTB_ERROR_UNEXPECTEDEOF;
    goto readfail;
  }
#ifdef __L_ENDIAN__
  Container->cachedHeader.fastDataMarker = rttb_Swap32(Container->cachedHeader.fastDataMarker);
#endif /*__L_ENDIAN__*/
  /*
    There now follows a table of N * QWORDS, these are the lengths of the
    files in the RTTB file, read them as DWORDs, skipping every other for
    future 64-bit support.  Unfortunately, I cannot handle large files
    (over 4 GB) yet, unless you're using a 64-bit processor.  In which case
    LONG_MAX will be > 3,000,000,000
  */

  c = Container->cachedHeader.n;
#if ( LONG_MAX <= 3000000000L )
  c *= 2; /* I have to do this to get the loop to work */
#endif

  for ( i = 0U; i < c; i++ ) {
#if ( LONG_MAX <= 3000000000L )
    if ( !( i % 2 ) ) { /* Every other */
      fseek(Container->boundfile.pfs, sizeof(WORD32), SEEK_CUR);
    }
    else {
#endif
      if ( fread(&wlong, sizeof(wlong), 1, Container->boundfile.pfs) < 1 ) {
        status = RTTB_ERROR_UNEXPECTEDEOF;
        goto readfail;
      }
#ifdef __L_ENDIAN__ /* Little endian machine */
      wlong = (long int)rttb_Swap32((WORD32)wlong);  /* FIXME: 64-bit processor would break here! */
#endif /*__L_ENDIAN__*/
      /* Add this w32 to the linked list of file lengths */
      if ( !Container->cachedHeader.lengths.Add(&Container->cachedHeader.lengths, &wlong) ) {
        status = RTTB_ERROR_MALLOC;
        goto readfail;
      }
#if ( LONG_MAX <= 3000000000L )
    }
#endif
  }
  /* Now read table of checksums of all files (DWORDs * N) */
  for ( i = 0U; i < Container->cachedHeader.n; i++ ) {
    if ( fread(&w32, sizeof(w32), 1, Container->boundfile.pfs) < 1 )
      return RTTB_ERROR_UNEXPECTEDEOF;
#ifdef __L_ENDIAN__
    w32 = rttb_Swap32(w32);
#endif /*__L_ENDIAN__*/
    /* Add this w32 to the linked list of checksums */
    if ( !Container->cachedHeader.checksums.Add(&Container->cachedHeader.checksums, &w32) )
      return RTTB_ERROR_MALLOC;
  }

  /* Now read the table of lengths of filenames (WORDs * N) */
  lenfns = (WORD16*)rttb_malloc( sizeof(WORD16) * (Container->cachedHeader.n) );
  if ( !lenfns ) {
    status = RTTB_ERROR_MALLOC;
    goto readfail;
  }

  if ( fread(lenfns, sizeof(WORD16), Container->cachedHeader.n, Container->boundfile.pfs) < Container->cachedHeader.n ) {
    rttb_free(lenfns);
    return RTTB_ERROR_UNEXPECTEDEOF;
  }
  /* Do swaps on filename lenghs if neccersary */
#ifdef __L_ENDIAN__
  for ( i = 0U; i < Container->cachedHeader.n; i++ ) {
    lenfns[i] = rttb_Swap16(lenfns[i]);
  }
#endif /*__L_ENDIAN__*/
  /* There is no linked list for the filename lengths, these are just to assist
  us is loading the filenames, which come next... */
  for ( i = 0U; i < (unsigned int)Container->cachedHeader.n; i++ ) {
    char* fn = rttb_malloc( (size_t)(sizeof(char)*(lenfns[i]+1)) ); /* Allocate enough space for current filename */
    if ( !fn ) {
      rttb_free(lenfns);
      status = RTTB_ERROR_MALLOC;
      goto readfail;
    }
    if ( fread(fn, (size_t)lenfns[i], 1, Container->boundfile.pfs) < 1 ) { /* Read in current filename from file */
      rttb_free(lenfns);
      rttb_free(fn);
      status = RTTB_ERROR_UNEXPECTEDEOF;
      goto readfail;
    }
    fn[lenfns[i]] = '\0'; /* Terminator fn for use of C string library */
    Container->cachedHeader.fileNames.Add(&Container->cachedHeader.fileNames, fn);
    rttb_free(fn);
  }

  rttb_free(lenfns);readfail:
  /* Clean up and get out of here */
  rewind(Container->boundfile.pfs);
  if ( stackedFilePos && stackedFilePos != -1L )
    fseek(Container->boundfile.pfs, stackedFilePos, SEEK_SET);
  return status;
}
/*-------------------------------------------------------------------------*/
static void _i_initheader(
  struct header* p
)
{
  if (p) {
    p->revision = RTTB_REVISION;
    p->n = (WORD32)0U;
    p->fastDataMarker = (WORD32)0U;
    rttb_link_ptrs_init(&p->lengths, sizeof(long int));
    rttb_link_ptrs_init(&p->checksums, sizeof(WORD32));
    rttb_internal_strzzConstruct(&p->fileNames);
  }
}
/*-------------------------------------------------------------------------*/
static void _i_freeheader(
  struct header* p
)
{
  if (p) {
    p->lengths.DelAll(&p->lengths);
    p->lengths.DelAll(&p->checksums);
    p->fileNames.Destruct(&p->fileNames);
    _i_initheader(p);
  }
}
/*-------------------------------------------------------------------------*/
static void _i_initrttbfile(
  struct rttbfile* p
)
{
  if (p) {
    p->rttbHandle = RTTB_NULLHANDLE;
    rttb_core_InitBoundFile(&p->boundfile);
    p->lenBytes = CACHE_INVALID;
    p->activeMarker = 0L; /* All files start at top as standard */
  }
}
/*-------------------------------------------------------------------------*/
static void _i_freerttbfile(
  struct rttbfile* p
)
{
  if (p) {
    _i_freeboundfile(&p->boundfile);
    _i_initrttbfile(p);
  }
}
/*-------------------------------------------------------------------------*/
static void _i_initrttbcontainer(
  struct rttbcontainer* p
)
{
  if (p) {
    unsigned int i;

    rttb_core_InitBoundFile(&p->boundfile);
    for ( i = 0U; i < RTTB_MAXOPEN_PERRTTB; i++ )
      _i_initrttbfile(&p->openFileArray[i]);

    _i_initheader(&p->cachedHeader);
    p->prepared = 0;
    p->lenBytes = CACHE_INVALID;
  }
}
/*-------------------------------------------------------------------------*/
static void _i_freerttbcontainer(
  struct rttbcontainer* p
)
{
  if (p) {
    unsigned int i;

    _i_freeboundfile(&p->boundfile);
    for ( i = 0U; i < RTTB_MAXOPEN_PERRTTB; i++ )
      _i_freerttbfile(&p->openFileArray[i]);

    _i_freeheader(&p->cachedHeader);
    _i_initrttbcontainer(p);
  }
}
/*-------------------------------------------------------------------------*/
void rttb_core_InitBoundFile(
  BOUNDFILE* p
)
{
  if (p) {
    p->pfs = NULL;
    p->filename = NULL;
    p->modes = NULL;
  }
}
/*-------------------------------------------------------------------------*/
static void _i_freeboundfile(
  BOUNDFILE* p
)
{
  if (p) {
    if (p->pfs) rttb_close(p->pfs);
    if (p->filename) rttb_free(p->filename);
    if (p->modes) rttb_free(p->modes);
    rttb_core_InitBoundFile(p);
  }
}
/*-------------------------------------------------------------------------*/
static void _i_releaseallhandles(
  struct rttbcontainer* p
)
{
  if (p) {
    unsigned int i;
    for ( i = 0U; i < RTTB_MAXOPEN_PERRTTB; i++ ) {
      if ( p->openFileArray[i].rttbHandle )
        _i_freerttbfile(&p->openFileArray[i]);
    }
  }
}
/*-------------------------------------------------------------------------*/
static int _i_checkrttbname(
  const char* Filename,
  unsigned int* PIndex
)
{
  if (Filename) {
    unsigned int i, c;

    c = _rttblist.Count(&_rttblist);
    for ( i = 0U; i < c; i++ ) {
      struct rttbcontainer container;

      _rttblist.Get(&_rttblist, i, &container);

      if ( container.boundfile.filename ) {
        if ( strcmp(container.boundfile.filename, Filename) == 0 ) {
          if (PIndex) *PIndex = i;
          return 1; /* TRUE; */
        }
      }
    }
  }
  return 0; /* No such rttb container */
}
/*-------------------------------------------------------------------------*/
static void _i_releaseallcontainers()
{
  while ( _rttblist.Count(&_rttblist) ) {
    struct rttbcontainer container;

    _rttblist.Get(&_rttblist, 0U, &container);
    _i_freerttbcontainer(&container);
    _rttblist.DelByPos(&_rttblist, 0U);
  }
  return;
}
/*--------------------------------------------------------------------------*/
static RTTB_ERROR rttb_core_PrepareFile(
  const char* RTTB,
  const char* FileName
)
{
  unsigned int i, c;
  unsigned int fnIndex; /* May not be neccersary */
  struct rttbcontainer container;
  bool detected = false; /* For missing RTTB detection and getting loop wrapped into one */
  struct rttbfile* Prttbfile;

  assert(_started);
  assert(RTTB);
  assert(FileName);
  /* Implictly trust callers in release mode */
  if ( !RTTB )
    return RTTB_ERROR_INVRTTB; /* No such container */

  /* Search for and pick out the container, if not there then set error
  the same as above, the get and set whether it's there or not is rolled
  into one as opposed to calling helper functions because I think it will
  be faster this way, although PrepareFile() probally isn't a critical
  function */
  c = _rttblist.Count(&_rttblist);
  for ( i = 0U; i < c; i++ ) {
    _rttblist.Get(&_rttblist, i, &container);
    if ( container.boundfile.filename ) {
      if ( !strcmp(container.boundfile.filename, RTTB) ) { /* Match? */
        detected = true;
        break; /* Don't waste time */
      }
    }
  }
  if ( !detected ) return RTTB_ERROR_INVRTTB; /* No such container */

  /* Search for file */
  fnIndex = container.cachedHeader.fileNames.Find(&container.cachedHeader.fileNames, FileName);
  if ( fnIndex == -1 ) /* File not found */
    return RTTB_ERROR_NOSUCHFILE;

  Prttbfile = _i_freefile(&container); /* Get the address of the first free file position (not index!) */
  if ( !Prttbfile )
    return RTTB_ERROR_OPENLIMIT;

  Prttbfile->rttbHandle = rttb_core_GenerateHandle();
  Prttbfile->boundfile.pfs = NULL; /* Contained files do not have their own file stream */
  assert(!Prttbfile->boundfile.filename); /* We should not be re-assigning a previous string */
  Prttbfile->boundfile.filename = rttb_malloc( sizeof(char)*(strlen(FileName)+1));
  if ( Prttbfile->boundfile.filename ) {
    strcpy(Prttbfile->boundfile.filename, FileName);
  }
  else {
    /* No need to cleanup, container not written back to list */
    return RTTB_ERROR_MALLOC; /* Out of memory */
  }
  /* Before I forget, write back the container to the list */
  _rttblist.Set(&_rttblist, i, &container);
  return RTTB_ERROR_NOERROR;
}
/*--------------------------------------------------------------------------*/
static int rttb_core_CheckFileExists(
  const char* RTTB,
  const char* FileName
)
{
  unsigned int i, fnCount;
  struct rttbcontainer container;

  assert(_started);
  if ( !_i_checkrttbname(RTTB, &i) )
    return 0;

  /* Now I have index of container in rttb list */
  _rttblist.Get(&_rttblist, i, &container);
  assert(container.prepared); /* Oops, container must be preared before this call */
  /* Reuse i to loop through container's header in memory looking for
  the specified filename */
  fnCount = container.cachedHeader.fileNames.Count(&container.cachedHeader.fileNames);
  for ( i = 0U ; i < fnCount; i++ ) {
    const char* fnCurrent = container.cachedHeader.fileNames.GetByIndex(&container.cachedHeader.fileNames, i);
    if ( fnCurrent ) {
      if ( !strcmp(fnCurrent, FileName) )
        return 1; /* File exists */
    }
  }
  return 0; /* File does not exist */
}
/*--------------------------------------------------------------------------*/
static int rttb_core_IsValidHandle(
  const RTTB_HANDLE Handle
)
{
  unsigned int i, c;

  assert(_started);
  c = _rttblist.Count(&_rttblist);
  for ( i = 0U; i < c; i++ ) {
    unsigned int j;
    struct rttbcontainer container;

    _rttblist.Get(&_rttblist, i, &container);
    for ( j = 0U; j < RTTB_MAXOPEN_PERRTTB; j++ ) {
      if ( container.openFileArray[j].rttbHandle == Handle )
        return 1;
    }
  }
  return 0;
}
/*--------------------------------------------------------------------------*/
static void rttb_core_SetFileMode(
  const RTTB_HANDLE RTTBHandle,
  const char* Mode
)
{
  /* We won't bother with any error checking, if the file's here, we'll set it */
  unsigned int i, c;

  assert(_started);
  c = _rttblist.Count(&_rttblist);
  for ( i = 0U; i < c; i++ ) {
    struct rttbcontainer container;
    unsigned int j;

    _rttblist.Get(&_rttblist, i, &container);
    for ( j = 0U; j < RTTB_MAXOPEN_PERRTTB; j++ ) {
      if ( container.openFileArray[j].rttbHandle == RTTBHandle ) {
        container.openFileArray[j].boundfile.modes = _i_setstr(container.openFileArray[j].boundfile.modes, Mode);
        _rttblist.Set(&_rttblist, i, &container);
        break;
      }
    }
  }
}
/*--------------------------------------------------------------------------*/
static const char *rttb_core_GetFileMode(
  const RTTB_HANDLE RTTBHandle
)
{
  unsigned int i, c;

  assert(_started);
  c = _rttblist.Count(&_rttblist);
  for ( i = 0U; i < c; i++ ) {
    struct rttbcontainer container;
    unsigned int j;
    _rttblist.Get(&_rttblist, i, &container);
    for ( j = 0U; j < RTTB_MAXOPEN_PERRTTB; j++ ) {
      if ( container.openFileArray[j].rttbHandle == RTTBHandle )
        return container.openFileArray[j].boundfile.modes;
    }
  }
  return NULL;
}
/*--------------------------------------------------------------------------*/
static char* _i_setstr(
  char *str,
  const char *NewString
)
{
  /* Release old */
  if ( str ) {
    rttb_free(str);
    str = NULL;
  }
  if ( NewString ) {
    if ( str ) {
      if ( !strcmp(str, NewString) ) /* Strings don't differ, don't waste time */
        return str; /* Pretend success */
    }
    str = rttb_malloc(sizeof(char)*(strlen(NewString)+1));
    if ( !str )
      return NULL;
    strcpy(str, NewString);
  }
  return str;
}
/*--------------------------------------------------------------------------*/
static struct rttbfile* _i_freefile(
  struct rttbcontainer* Container
)
{
  unsigned int i;
  for ( i = 0U; i < RTTB_MAXOPEN_PERRTTB; i++ ) {
    if ( Container->openFileArray[i].rttbHandle == RTTB_NULLHANDLE )
      return &Container->openFileArray[i];
  }
  return NULL;
}
/*--------------------------------------------------------------------------*/
static unsigned int _i_assosiationcount(
  struct rttbcontainer* Container
)
{
  size_t i;
  unsigned int c = 0U;

  for ( i = 0U; i < RTTB_MAXOPEN_PERRTTB; i++ ) {
    if ( Container->openFileArray[i].rttbHandle != RTTB_NULLHANDLE )
      c++;
  }
  return c;
}
/*-------------------------------------------------------------------------*/
static RTTB_HANDLE rttb_core_GetHandle(
  const char *RTTB,
  const char *FileName
)
{
  unsigned int i, c;
  assert(_started);
  c = _rttblist.Count(&_rttblist);
  for ( i = 0U; i < c; i++ ) {
    struct rttbcontainer container;

    _rttblist.Get(&_rttblist, i, &container);
    assert(container.boundfile.filename);
    if ( strcmp(container.boundfile.filename, RTTB) == 0 ) {
      /* This is the container in question */
      unsigned int j;
      for ( j = 0U; j < RTTB_MAXOPEN_PERRTTB; j++ ) {
        if ( strcmp(container.openFileArray[j].boundfile.filename, FileName) == 0 ) {
          /* This is the one */
          return container.openFileArray[j].rttbHandle;
        }
      }
      break; /* RTTB matched but filename not?  Bad then, con't bother going on */
    }
  }
  return RTTB_NULLHANDLE;
}
/*-------------------------------------------------------------------------*/
static RTTB_ERROR rttb_core_GetFileOffset(
  const RTTB_HANDLE Handle,
  long int* POffset
)
{
  unsigned int i, c;

  assert(Handle != RTTB_NULLHANDLE); /* Somebody set up us the bomb! */
  c = _rttblist.Count(&_rttblist);
  for ( i = 0U; i < c; i++ ) {
    struct rttbcontainer container;
    unsigned int j;

    _rttblist.Get(&_rttblist, i, &container); /* Get a container */
    for ( j = 0U; j < RTTB_MAXOPEN_PERRTTB; j++ ) {
      if ( container.openFileArray[j].rttbHandle == Handle ) {
        bool modified;
        long int offset;

        if ( container.openFileArray[j].boundfile.pfs ) /* Loose file! */
          offset = _i_loose_getfileoffset(&container, j, &modified);
        else /* Contained file */
          offset = _i_contained_getfileoffset(&container, j, &modified);

        if ( POffset ) *POffset = offset; /* Caller wants offset?  Give it to them */

        /* If the offset was modified the container must be written back into
        the master rttblist.  In case you don't understand why it might be
        modified in a getter it's because the cache can be updated therefore
        written to. */
        if ( modified )
          _rttblist.Set(&_rttblist, i, &container);
        return RTTB_ERROR_NOERROR; /* Success */
      }
    }
  }
  return RTTB_ERROR_INVHANDLE; /* Invalid handle */
}
/*-------------------------------------------------------------------------*/
static RTTB_ERROR rttb_core_SetFileOffset(
  const RTTB_HANDLE Handle,
  const long int NewOffsetRecord
)
{
  /* Attention: This function does not seek!  That's up to rttb_fseek.c,
  this function merely updates the records of such seeks. */
  unsigned int i, c;

  assert(Handle != RTTB_NULLHANDLE); /* Somebody set up us the bomb! */
  c = _rttblist.Count(&_rttblist);
  for ( i = 0U; i < c; i++ ) {
    struct rttbcontainer container;
    unsigned int j;

    _rttblist.Get(&_rttblist, i, &container); /* Get a container */
    for ( j = 0U; j < RTTB_MAXOPEN_PERRTTB; j++ ) {
      if ( container.openFileArray[j].rttbHandle == Handle ) {
        /* I happen to know that _i_loose/contained_setfileoffset are the same
        thing ;-D so I'll save code and use either wihtout a care */
        _i_contained_setfileoffset(&container, j, NewOffsetRecord);
      }
      _rttblist.Set(&_rttblist, i, &container); /* Writeback to master container list */
      return RTTB_ERROR_NOERROR;
    }
  }
  return RTTB_ERROR_INVHANDLE; /* Invalid handle */
}
/*-------------------------------------------------------------------------*/
static RTTB_ERROR rttb_core_GetFileLength(
  const RTTB_HANDLE Handle,
  long int* PLength
)
{
  unsigned int i, c;

  c = _rttblist.Count(&_rttblist);
  for ( i = 0U; i < c; i++ ) { /* All containers */
    struct rttbcontainer container;
    unsigned int j;

    _rttblist.Get(&_rttblist, i, &container); /* Get a container */
    for ( j = 0U; j < RTTB_MAXOPEN_PERRTTB; j++ ) {
      if ( container.openFileArray[i].rttbHandle == Handle ) { /* Found the file in question */
        long int length;
        bool modified;

        if ( container.openFileArray[i].boundfile.pfs ) /* Loose file! */
          length = _i_loose_getfilelength(&container, j, &modified);
        else /* Contained */
          length = _i_contained_getfilelength(&container, j, &modified);

        if ( PLength ) *PLength = length; /* Update the caller if they want it (which they should) */

        /* If the length cache was modified writeback the container to the masterlist */
        if ( modified )
          _rttblist.Set(&_rttblist, i, &container);
        return RTTB_ERROR_NOERROR;
      }
    }
  }
  return RTTB_ERROR_INVHANDLE; /* Invalid handle */
}
/*-------------------------------------------------------------------------*/
static size_t rttb_core_ContainedRead(
  RTTB_HANDLE Handle,
  size_t Size,
  size_t N,
  void* PTarget,
  RTTB_ERROR* PError
)
{
  unsigned int i, c;
  size_t n_read = 0U; /* How many we successfully read */

  if ( PError )
    *PError = RTTB_ERROR_NOERROR;

  c = _rttblist.Count(&_rttblist);
  for ( i = 0U; i < c; i++ ) { /* All containers */
    struct rttbcontainer container;
    unsigned int j;

    _rttblist.Get(&_rttblist, i, &container);
    for ( j = 0U; j < RTTB_MAXOPEN_PERRTTB; j++ ) {
      if ( container.openFileArray[j].rttbHandle == Handle ) { /* This is the one we're after */
        RTTB_ERROR tError;
        size_t Ncount;
        char* destination = (char*)PTarget;
        RTTB_UINT headerIndex;
        long int off; /* Offset of file in container */

        assert(!container.openFileArray[j].boundfile.pfs); /* No!  No fscking loose files get into this function! */

        headerIndex = _i_OpenArrayToHeaderIndex(&container, j);
                  off = _i_LocalToGlobalOffset(
                         &container,
                         headerIndex,
          _i_GetFileOffInCacheHead(&container, headerIndex) /* File must be open and in the open file array for this to work or it returns 0 */
                  );
        tError = _i_container_seek(&container, off);
        if ( tError != RTTB_ERROR_NOERROR ) {
          if ( PError ) *PError = tError;
          break; /* Should break out of complete nest : DEBUG to check! */
        }
        Ncount = N;
        while ( Ncount-- ) {
          tError = _i_directread(container.boundfile.pfs, Size, destination);
          if ( tError != RTTB_ERROR_NOERROR ) {
            if ( PError ) *PError = tError;
            break; /* Should break out of complete nest : DEBUG to check! */
          }
          destination += Size; /* Move to next chunk */
          n_read++; /* Successfully read a block */
        }
        break; /* Don't try more handles */
      }
    }
    _i_contained_setfileoffset(&container, j, _i_contained_getfileoffset(&container, i, NULL) + (Size * N) ); /* Update offset record */
    _rttblist.Set(&_rttblist, i, &container); /* Offset may have changed */
  }
  return n_read;
}
/*-------------------------------------------------------------------------*/
static long int rttb_core_LocalToGlobalOffsetByHandle(
  RTTB_HANDLE Handle,
  long int LocalOffset
)
{
  unsigned int i, c;
  long int global = -1L;

  c = _rttblist.Count(&_rttblist);
  for ( i = 0U; i < c; i++ ) { /* Containers */
    struct rttbcontainer container;
    unsigned int j;

    _rttblist.Get(&_rttblist, i, &container);
    for ( j = 0U; j < RTTB_MAXOPEN_PERRTTB; j++ ) {
      if ( container.openFileArray[j].rttbHandle == Handle ) {
        global = _i_LocalToGlobalOffset(&container, j, LocalOffset);
        break;
      }
    }
  }
  return global;
}
/*-------------------------------------------------------------------------*/
static long int _i_LocalToGlobalOffset(
  struct rttbcontainer* Container,
  RTTB_UINT Index,
  long int LocalOffset
)
{
  unsigned int i, c; /* Linked list transcender loop */
  long int global = _i_getfastdata(Container);

  if ( global == -1L )
    return global;
  i = Index;

  c = Container->cachedHeader.lengths.Count(&Container->cachedHeader.lengths);
  while ( i-- && c-- ) {
    long int length;

    Container->cachedHeader.lengths.Get(&Container->cachedHeader.lengths, i, &length);
    global += length;
  }
  global += LocalOffset;
  return global;
}
/*-------------------------------------------------------------------------*/
static long int _i_loose_getfileoffset(
  struct rttbcontainer* PContainer,
  const unsigned int Index,
  bool *PModFlag
)
{
  assert(Index < RTTB_MAXOPEN_PERRTTB); /* If not so, we're fux0red! */
  if ( PModFlag ) *PModFlag = false;
  if ( PContainer->openFileArray[Index].activeMarker == CACHE_INVALID ) { /* Not cached? */
    long int tl;
    tl = ftell(PContainer->openFileArray[Index].boundfile.pfs);
    if ( tl != -1L ) { /* No error */
      PContainer->openFileArray[Index].activeMarker = tl; /* Cache */
      if ( PModFlag )
        *PModFlag = true;
    }
  }
  return PContainer->openFileArray[Index].activeMarker;
}
/*-------------------------------------------------------------------------*/
static long int _i_contained_getfileoffset(
  struct rttbcontainer* PContainer,
  const unsigned int Index,
  bool* PModFlag
)
{
  assert(Index < RTTB_MAXOPEN_PERRTTB); /* Index must be within range to prevent crash */
  /* Note: Cache just dummy here, cache is primary copy, caller needs to know
  if we touched it for _rttblist writeback reasons, we do not. */
  if ( PModFlag ) *PModFlag = false;
  return PContainer->openFileArray[Index].activeMarker;
}
/*-------------------------------------------------------------------------*/
static long int _i_container_getfileoffset(
  struct rttbcontainer* PContainer,
  bool* PModFlag
)
{
  if ( PModFlag ) *PModFlag = false; /* Cache no longer exists because I had to get rid of it, never any mods */
  assert(PContainer->boundfile.pfs);
  assert(PContainer->prepared);
  return ftell(PContainer->boundfile.pfs);
}
/*-------------------------------------------------------------------------*/
static void _i_loose_setfileoffset(
  struct rttbcontainer* PContainer,
  const unsigned int Index,
  const long int Offset
)
{
  assert(Index < RTTB_MAXOPEN_PERRTTB);
  PContainer->openFileArray[Index].activeMarker = Offset;
}
/*-------------------------------------------------------------------------*/
static long int _i_loose_getfilelength(
  struct rttbcontainer* PContainer,
  const unsigned int Index,
  bool* PModFlag
)
{
  assert(Index < RTTB_MAXOPEN_PERRTTB); /* Index must be in range */
  if ( PModFlag ) *PModFlag = false; /* Cache not modified yet... */
  if ( PContainer->openFileArray[Index].lenBytes == CACHE_INVALID ) { /* Not cached? */
    long int li;

    li = _i_getphysicallength(PContainer->openFileArray[Index].boundfile.pfs);
    if ( li != -1L ) { /* No error */
      PContainer->openFileArray[Index].lenBytes = li; /* Cache */
      if ( PModFlag )
        *PModFlag = true; /* Had to modify the cache, tell caller so they can writeback the container */
    }
  }
  return PContainer->lenBytes;
}
/*-------------------------------------------------------------------------*/
static long int _i_contained_getfilelength(
  struct rttbcontainer* PContainer,
  const unsigned int Index,
  bool* PModFlag
)
{
  assert(Index < RTTB_MAXOPEN_PERRTTB); /* Index must be within range to prevent crash */
  /* Note: Cache just dummy here, cache is primary copy, caller needs to know
  if we touched it for _rttblist writeback reasons */
  if ( PModFlag ) *PModFlag = false;
  if ( PContainer->openFileArray[Index].lenBytes == CACHE_INVALID ) {
    /* Cache miss, OK the fast cache of the length is invalid but we
    won't waste much time getting it from the cached header linked list.
    First though, get the header entry index for the open file index */
    RTTB_UINT headerIndex = _i_OpenArrayToHeaderIndex(PContainer, Index);
    PContainer->cachedHeader.lengths.Get(
      &PContainer->cachedHeader.lengths,  /* (this) < for C++ programmers */
      headerIndex,                        /* Where we know it's at in the linked list */
      &PContainer->openFileArray[Index].lenBytes   /* Get data into cache */
    );
    if ( PModFlag )
      *PModFlag = true;
  }
  return PContainer->openFileArray[Index].lenBytes;
}
/*-------------------------------------------------------------------------*/
static long int _i_container_getfilelength(
  struct rttbcontainer* PContainer,
  bool* PModFlag
)
{
  if ( PModFlag ) *PModFlag = false; /* Init the modify flag so it's false isf not true */
  if ( PContainer->lenBytes == CACHE_INVALID ) { /* Not cached? */
    long int fl;
    fl = _i_getphysicallength(PContainer->boundfile.pfs);
    if ( fl != -1L ) { /* No error */
      PContainer->lenBytes = fl; /* Cache so we don't have to call tl again */
      if ( PModFlag ) /* Caller wan't notification of modification? */
        *PModFlag = true; /* Do so */
    }
  }
  return PContainer->lenBytes; /* Return from cache */
}
/*-------------------------------------------------------------------------*/
static long int _i_getphysicalpos(
  struct rttbcontainer* Container,
  const RTTB_UINT MasterListIndex,
  const RTTB_UINT Index,
  bool* PModFlag
)
{
  long int offset;
  unsigned long i;
  bool modFlag = false; /* This is used incase the caller does not hold the most recent container and expects us to do the writeback */

  assert( Index <= RTTB_MAXOPEN_PERRTTB );

  /* This probally needs a little caching which should be a linked list
  in the rttbcontainer with the start offsets... but would it be faster?! */
  offset = _i_getfastdata(Container); /* Base point - start of raw files */
  for ( i = 0U; i < Index; i++ )
    offset += _i_contained_getfilelength(Container, i, PModFlag);

  if ( modFlag ) {
    /* We can't writeback the container if the caller holds a copy or
    theirs would be invalidated, if they have not passed a pointer
    we know it's safe to do a writeback */
    if ( PModFlag )
      *PModFlag = modFlag;
    else
      _rttblist.Set(&_rttblist, MasterListIndex, Container);
  }
  return offset;
}
/*-------------------------------------------------------------------------*/
static inline void _i_recover(
  FILE* fsp,
  const long int OldPos
)
{
  rewind(fsp);
  fseek(fsp, OldPos, SEEK_SET);
}
/*-------------------------------------------------------------------------*/
static long int _i_getphysicallength(
  FILE* fsp
)
{
  long int oldPos, length;

  oldPos = ftell(fsp); /* Save offset for recovery */
  if ( oldPos == -1L ) return oldPos;
  fseek(fsp, 0, SEEK_END); /* Go to the end */
  length = ftell(fsp); /* Save end position (is this out by one of length?  DEBUG) */

  _i_recover(fsp, oldPos);
  return length;
}
/*-------------------------------------------------------------------------*/
static RTTB_ERROR _i_directread(
  FILE* FileStream,
  size_t Bytes,
  void* RawData
)
{
  RTTB_ERROR err = RTTB_ERROR_NOERROR;

  if ( FileStream && RawData && Bytes ) {
    if ( fread(RawData, Bytes, 1, FileStream) < 1 )
      err = RTTB_ERROR_EOF;

    return err;
  }
  return RTTB_ERROR_INVPARAMS; /* NULL pointer(s) */
}
/*-------------------------------------------------------------------------*/
static inline long int _i_getfastdata(
  struct rttbcontainer* Container
)
{
  return (long int)Container->cachedHeader.fastDataMarker;
}
/*-------------------------------------------------------------------------*/
static RTTB_ERROR _i_container_seek(
  struct rttbcontainer* Container,
  long int Offset
)
{
  long int savedOffset; /* Seek back incase of trouble */

  /* Must be a valid, prepared container */
  assert(Container->boundfile.pfs); /* File open */
  assert(Container->prepared); /* Contents read properly from file into caches */

  /* Check for seek past end to save time calling I/O functions */
  if ( Offset >= _i_container_getfilelength(Container, NULL) )
    return RTTB_ERROR_EOF;

  savedOffset = _i_container_getfileoffset(Container, NULL);
  if ( Offset != savedOffset )
    fseek(Container->boundfile.pfs, Offset, SEEK_SET);
  /* Note: DOS can fail and report seeks are OK when not, to verify we
  must manually ftell(), this is really a limitation aided and abetted
  by Borland who should call ftell() and verify it before returning from
  fseek(). */
  if ( ftell(Container->boundfile.pfs) != Offset ) {
    _i_recover(Container->boundfile.pfs, savedOffset);
    return RTTB_ERROR_UNEXPECTED;
  }
  return RTTB_ERROR_NOERROR;
}
/*-------------------------------------------------------------------------*/
static RTTB_UINT _i_OpenArrayToHeaderIndex(
  struct rttbcontainer* Container,
  RTTB_UINT Index
)
{
  RTTB_UINT i, c;
  bool matched = false;
  assert(Container);
  assert(Index < RTTB_MAXOPEN_PERRTTB);

  assert( !(Container->openFileArray[Index].boundfile.pfs) ); /* This function only to be used on contained files!! */

  c = Container->cachedHeader.fileNames.Count(&Container->cachedHeader.fileNames); /* c = number of files total in container */
  /* Limiters */
  if ( Index > RTTB_MAXOPEN_PERRTTB )
    Index = RTTB_MAXOPEN_PERRTTB;
  if ( Index > c )
    Index = c;

  for ( i = 0U; i < c; i++ ) { /* Loop through actual filename entries in the RTTB container */
    const char* fnptr; /* Filename pointer ( from linked list ) */

    fnptr = _i_GetFNPtrInCacheHead(Container, i);
    if ( fnptr && Container->openFileArray[Index].boundfile.filename ) {
      if ( !strcmp(Container->openFileArray[Index].boundfile.filename, fnptr ) ) {
        /* This is the file, compared by name, i must be the header index
        of the equivilant said Index */
        matched = true;
        break;
      }
    }
  }
  if ( !matched ) return 0U;
  return i;
}
/*-------------------------------------------------------------------------*/
static RTTB_UINT _i_HeaderIndexToOpenArray(
  struct rttbcontainer* Container,
  RTTB_UINT HeaderIndex
)
{
  RTTB_UINT i;
  bool matched = false;
  const char* fnptr;

  assert(Container);

  fnptr = _i_GetFNPtrInCacheHead(Container, HeaderIndex);
  for ( i = 0U; i < RTTB_MAXOPEN_PERRTTB; i++ ) {
    if ( Container->openFileArray[i].boundfile.filename ) {
      if ( !strcmp(Container->openFileArray[i].boundfile.filename, fnptr) ) {
        /* This is the file, compared by name, i must be the open file array
        index of the equivilant said header index */
        matched = true;
        break;
      }
    }
  }
  if ( !matched ) return 0U;
  return i;
}
/*-------------------------------------------------------------------------*/
static long int _i_GetFileOffInCacheHead(
  struct rttbcontainer* Container,
  RTTB_UINT HeaderIndex
)
{
  RTTB_UINT oai = _i_HeaderIndexToOpenArray(Container, HeaderIndex); /* Get index for addressing open array */
  return Container->openFileArray[oai].activeMarker;
}
/*-------------------------------------------------------------------------*/
static inline const char* _i_GetFNPtrInCacheHead(
  struct rttbcontainer* Container,
  RTTB_UINT HeaderIndex
)
{
  return Container->cachedHeader.fileNames.GetByIndex(&Container->cachedHeader.fileNames, HeaderIndex);
}
/*-------------------------------------------------------------------------*/
static RTTB_BYTE rttb_core_ReadOneByte(
  RTTB_HANDLE Handle,
  RTTB_ERROR* PError
)
{
  UINT i, c;
  RTTB_BYTE onebyte;
  RTTB_ERROR tError = RTTB_ERROR_NOERROR;

  if ( PError )
    *PError = tError;

  c = _rttblist.Count(&_rttblist);
  for ( i = 0U; i < c; i++ ) {
    struct rttbcontainer container;
    UINT j;

    _rttblist.Get(&_rttblist, i, &container);
    for ( j = 0U; j < RTTB_MAXOPEN_PERRTTB; j++ ) {
      if ( container.openFileArray[j].rttbHandle == Handle ) {
        UINT headerIndex;
        long int off; /* Offset of file in container (global) */

        assert(!container.openFileArray[j].boundfile.pfs); /* I'll add support for these later, shouldn't be hard */

        headerIndex = _i_OpenArrayToHeaderIndex(&container, j);
        off = _i_LocalToGlobalOffset(
          &container,
          headerIndex,
          _i_GetFileOffInCacheHead(&container, headerIndex)
        );
        tError = _i_container_seek(&container, off);
        if ( tError != RTTB_ERROR_NOERROR )
          break;
        /* Now I'm in the right place, do the byte read */
        onebyte = _i_readbyte(container.boundfile.pfs, &tError); /* Only works for contained file for now */
        if ( tError != RTTB_ERROR_NOERROR )
          break;
        /* Success, move the file's offset on a byte for successive calls */
        container.openFileArray[j].activeMarker++; /* For this file */

        /* Time to writeback the dirty container */
        _rttblist.Set(&_rttblist, i, &container);
      } /* Handle matched */ 
    } /* j loop */
    break; /* Don't try more containers for handles */
  } /* i loop */

  if ( PError )
    *PError = tError;
  return onebyte;
}
/*-------------------------------------------------------------------------*/
static char* rttb_core_ReadString(
  RTTB_HANDLE Handle,
  char* s,
  unsigned int size,
  RTTB_ERROR* PError
)
{
  /* FIXME: This is very inefficient.  It's a quick hack to avoid complexities */

  RTTB_BYTE oneByte;
  unsigned int i;

  if ( size ) {
    for ( i = 0U; i < size-1; i++ ) {
      oneByte = rttb_core_ReadOneByte(Handle, PError);
      if ( *PError == RTTB_ERROR_EOF ) {
        if ( i == 0 ) /* First read returned EOF? */
          s = NULL; /* Nothing can be read */
        break;
      }
      s[i] = oneByte;
    }
    s[i] = '\0';
  }
  else
    s = NULL;

  return s;
}
/*-------------------------------------------------------------------------*/
static RTTB_BYTE _i_readbyte(
  FILE* FileStream,
  RTTB_ERROR* PError
)
{
  /* Very similar but more efficient than _i_directread */
  int onebyte = 0;

  if ( PError ) *PError = RTTB_ERROR_NOERROR;

  if ( FileStream ) {
    onebyte = fgetc(FileStream);
    if ( onebyte == EOF ) {
      if ( PError )
        *PError = RTTB_ERROR_EOF;
      return '\0';
    }
  }
  return (RTTB_BYTE)onebyte;
}
/*-------------------------------------------------------------------------*/
