/*
Daybo Logic Run-Time TarBall Library And Tools
Copyright (c) 2000-2006, David Duncan Ross Palmer, Daybo Logic
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
      
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
      
    * Neither the name of the Daybo Logic nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

/*
  This module is purely for checking leaks of file handles and blocks of
  memory.  When librttb is built with NDEBUG defined this module will not
  even be linked by the makefile.  It contains no useful information to RTTB.
  and some of the code here is very damaging to performance because of para-
  noid checking.
*/

#define RTTB_SOURCE

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif /*HAVE_CONFIG_H*/

#include <time.h>
#include <stdio.h>
#include <assert.h>

#ifdef HAVE_STDLIB_H
# include <stdlib.h>
#endif /*HAVE_STDLIB_H*/

#ifdef HAVE_LIMITS_H
# include <limits.h>
#endif /*HAVE_LIMITS_H*/

#ifdef HAVE_STRING_H*/
# include <string.h>
#endif /*HAVE_STRING_H*/

#ifdef RTTB_HDRSTOP
# pragma hdrstop
#endif /*RTTB_HDRSTOP*/

#include "rttb.h"
#include "rttb_int.h"
#include "rttb_ll.h"

#ifdef DEBUG /* Whole module only compiled if debugging */
/*-------------------------------------------------------------------------*/
static struct link_ptrs _handles; /* Linked list of file handles */
static int _started = 0;
static unsigned long _peakBlocks = 0UL; /* The peak number of blocks allocated during execution */
static unsigned long _mallocCalls = 0UL; /* Number of calls to rttb_malloc() */
static unsigned long _freeCalls = 0UL; /* Number of calls to rttb_free() */
static struct {
  void* ptr; /* A record of a dynamic pointer in use by the library */
  unsigned int line; /* The line number which allocated that memory */
  char file[25]; /* Just enough space of a filename which allocates that block too */
  size_t bytes; /* Size in bytes of the chunk */
} _chunks[256]; /* Maximum amount of pointers we can track */

static unsigned int i_GarbageCollection(void); /* Frees all current handles, returns amount freed */
static int i_IsPresent(const FILE* Handle);
static int i_regpfile(FILE* NewFileHandle);
static void i_unregpfile(FILE* OldFileHandle);
static int i_RegisterPointer(void* Pointer); /* Returns index in _ptrs or -1 if can't be registered */
static void i_UnregisterPointer(void* Pointer);
static int i_IsPointerRegistered(void* Pointer); /* Returns index or -1 if not registered */
static void i_CheckNullPointer(void* Pointer); /* If Pointer == NULL, crash */
static void i_InitChunkInfo(const unsigned int Index);
static unsigned long i_CountChunks(void); /* Returns the number of chunks allocate to the library */
/*-------------------------------------------------------------------------*/
int rttb_debug_FileManagerStartup()
{
  if (!_started)
  {
    rttb_link_ptrs_init(&_handles, sizeof(FILE*));
    _started = 1;
    return _started; /*TRUE*/
  }
  return 0;
}
/*-------------------------------------------------------------------------*/
int rttb_debug_FileManagerIsStarted()
{
  return _started;
}
/*-------------------------------------------------------------------------*/
void rttb_debug_FileManagerShutdown()
{
  if (_started)
  {
    unsigned int garbCount;

    _started = 0;
    garbCount = i_GarbageCollection();
    if (garbCount)
      fprintf(RTTB_DEV_ERROR, "librttbd: Internally failed to free %u file handles!\n", garbCount);

    _handles.DelAll(&_handles); /* Cleanup the ol' link list */
  }
}
/*-------------------------------------------------------------------------*/
static unsigned int i_GarbageCollection(void)
{
  unsigned int count = 0U;
  if (_started)
  {
    unsigned int i, total;

    total = _handles.Count(&_handles);
    for ( i = 0U; i < total; i++ )
    {
      FILE* activeHandle;

      _handles.Get(&_handles, i, &activeHandle);
      if (activeHandle)
      {
        /* Found one unfreed handle */
        count++;
        fclose(activeHandle);
      }
    }
  }
  return count;
}
/*-------------------------------------------------------------------------*/
static int i_IsPresent(const FILE* Handle)
{
  unsigned int i, total;

  total = _handles.Count(&_handles);
  for ( i = 0U; i < total; i++ )
  {
    FILE* activeHandle;

    _handles.Get(&_handles, i, &activeHandle);
    if ( activeHandle == Handle )
      return 1; /* TRUE */
  }
  return 0; /* FALSE */
}
/*-------------------------------------------------------------------------*/
static int i_regpfile(FILE* NewFileHandle)
{
  if (_started)
  {
    if ( i_IsPresent(NewFileHandle) )
      return 0; /* Duplicate additions are NOT allowed! */

    _handles.Add(&_handles, &NewFileHandle);
    return 1;
  }
  return 0;
}
/*-------------------------------------------------------------------------*/
static void i_unregpfile(FILE* OldFileHandle)
{
  if (_started)
  {
    unsigned int i, total;

    total = _handles.Count(&_handles);
    for ( i = 0U; i < total; i++ )
    {
      FILE* fhCompare;

      _handles.Get(&_handles, i, &fhCompare);
      if ( fhCompare == OldFileHandle ) /* Found it */
      {
        _handles.DelByPos(&_handles, i);
        break; /* Since it's illegal to have duplicates, give up */
      }
    }
  }
}
/*-------------------------------------------------------------------------*/
static int i_RegisterPointer(void* Pointer)
{
  size_t i;

  if ( i_IsPointerRegistered(Pointer) != -1) {
    /* Oops!  Not supposed to be registered */
    rttb_debug_dbgmsg("Pointer already registered");
    abort();
  }
  for ( i = 0U; i < sizeof(_chunks)/sizeof(_chunks[0]); i++ ) {
    if ( !_chunks[i].ptr ) { /* Found blank space */
      _chunks[i].ptr = Pointer;
      return i;
    }
  }
  rttb_debug_dbgmsg("%p - Can\'t register, increase array _ptrs");
  return -1; /* No free space */
}
/*-------------------------------------------------------------------------*/
static void i_UnregisterPointer(void* Pointer)
{
  int index;

  index = i_IsPointerRegistered(Pointer);
  if ( index == -1 ) {
    rttb_debug_dbgmsg("Freed a non-existant pointer");
    return;
  }
  i_InitChunkInfo(index);
  return;
}
/*-------------------------------------------------------------------------*/
static int i_IsPointerRegistered(void* Pointer)
{
  size_t i;

  i_CheckNullPointer(Pointer);
  for ( i = 0U; i < sizeof(_chunks)/sizeof(_chunks[0]); i++ ) {
    if ( _chunks[i].ptr == Pointer )
      return i;
  }
  return -1;
}
/*-------------------------------------------------------------------------*/
static void i_CheckNullPointer(void* Pointer)
{
  if ( !Pointer ) {
    rttb_debug_dbgmsg("NULL pointer!  Set a breakpoint and trace the callstack, caller: %p");
    assert(Pointer);
    abort();
  }
}
/*-------------------------------------------------------------------------*/
static void i_InitChunkInfo(const unsigned int Index)
{
  assert(Index < sizeof(_chunks)/sizeof(_chunks[0])); /* Chunk valid #? */
  _chunks[Index].ptr = NULL;
  _chunks[Index].line = 0U;
  _chunks[Index].file[0] = '\0'; /* Termination of string at beginning */
  _chunks[Index].bytes = (size_t)0U;
  return;
}
/*-------------------------------------------------------------------------*/
FILE* rttb_debug_open(const char* path, const char* mode)
{
  FILE* yourstream;

  assert(path && mode);
  yourstream = fopen(path, mode);
  if (!yourstream) return yourstream;

  if ( !i_regpfile(yourstream) )
  {
    /* Can't add to link list, so be clean and falsify the error */
    fclose(yourstream);
    return NULL;
  }
  return yourstream;
}
/*-------------------------------------------------------------------------*/
int rttb_debug_close(FILE* stream)
{
  /* This code's not monster-fast, doesn't matter this whole module is just bloating debug code anyway */
  assert(i_IsPresent(stream));
  i_unregpfile(stream);
  return fclose(stream);
}
/*-------------------------------------------------------------------------*/
void* rttb_debug_malloc(size_t size, const unsigned int Line, const char* File)
{
  void* ptr;
  unsigned int registeredIndex;
  unsigned long count;

  _mallocCalls++;
  assert(size);
  ptr = malloc(size); /* Attempt allocation */
  if (!ptr) {
    fprintf(RTTB_DEV_ERROR,
            "Out of memory (debug build librttbd), location of requester:\n"
            "Line: %u, file: %s, size in bytes of block: %u\n",
            Line, File, size
    );
    abort();
  }
  /* Register it */
  registeredIndex = i_RegisterPointer(ptr);
  if ( registeredIndex == -1 ) {
    /* Can't register, clean up */
    free(ptr);
    return NULL;
  }
  /* Add extra debugging information to the chunk info */
  _chunks[registeredIndex].line = Line; /* Save line of requester */
  strncpy(_chunks[registeredIndex].file, File, sizeof(_chunks[0].file)/sizeof(_chunks[0].file[0])); /* Save filename of requester */
  _chunks[registeredIndex].bytes = size; /* Save size of chunk */

  count = i_CountChunks();
  if ( count > _peakBlocks) _peakBlocks = count; /* Update peak stat */
  return ptr;
}
/*-------------------------------------------------------------------------*/
void rttb_debug_free(void* p)
{
  _freeCalls++;
  i_UnregisterPointer(p);
  free(p);
}
/*-------------------------------------------------------------------------*/
unsigned long rttb_debug_DebugBlocks()
{
  unsigned long blocks = i_CountChunks();

  fprintf(stdout, "librttbd: David's debug output for debug builds, to get rid of\n"
                  "this use release build:\n\n"
                  "Peak blocks: %lu, blocks now: %lu <- should be zero!\n"
                  "%lu calls to malloc(), %lu calls to free() (balanced?)\n",
                  _peakBlocks, blocks, _mallocCalls, _freeCalls
  );
  return blocks;
}
/*-------------------------------------------------------------------------*/
void rttb_debug_dbgmsg(const char* message)
{
  if (message) fprintf(stdout, "librttbd (debug message) - %s\n", message);
}
/*-------------------------------------------------------------------------*/
void rttb_debug_chunkdump()
{
  FILE* handle;

  handle = fopen("rttbchunk.log", "wt");
  if ( handle ) {
    unsigned int i;
    unsigned int count = 0U; /* Count of chunks */
    unsigned long charge = 0UL; /* Total bytes */

    printf("RTTB is dumping chunk info...\n");
    fprintf(handle, "RTTB debug chunk information dump!  time:%lu\n", time(NULL));
    fprintf(handle, "ID address size  - owner file, line\n\n");
    for ( i = 0U; i < sizeof(_chunks)/sizeof(_chunks[0]); i++ ) {
      if ( _chunks[i].ptr ) {
        fprintf(handle, "%u %p %u  - %s L%u\n",
          i, _chunks[i].ptr, _chunks[i].bytes, _chunks[i].file, _chunks[i].line
        );
        count++;
        charge += (unsigned long)_chunks[i].bytes;
      }
    }
    fprintf(handle, "\nchunk count: %u, total charge: %lu\n", count, charge);
    fclose(handle);
  }
  return;
}
/*-------------------------------------------------------------------------*/
static unsigned long i_CountChunks(void)
{
  unsigned int i;
  unsigned long c = 0UL;

  for ( i = 0U; i < sizeof(_chunks)/sizeof(_chunks[0]); i++ ) {
    if ( _chunks[i].ptr )
     c++;
  }
  return c;
}
/*-------------------------------------------------------------------------*/
#else /* Not debug module */
/* ANSI C does not allow a module with no externals so this
can't be complately blank.  So I'll add a dummy variable here,
if possible try to excluse rttb_debug.c from being compiled and
keep the entire module out in the makefiles. */
extern int rttb_debug__dummyBlah = 0;
#endif /*DEBUG*/   /* Entire module debug builds only */
