/*
Daybo Logic Run-Time TarBall Library And Tools
Copyright (c) 2000-2006, David Duncan Ross Palmer, Daybo Logic
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
      
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
      
    * Neither the name of the Daybo Logic nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

#define RTTB_SOURCE

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif /*HAVE_CONFIG_H*/

#include <stdio.h>
#include <assert.h>

#ifdef HAVE_STDDEF_H
# include <stddef.h>
#endif /*HAVE_STDDEF_H*/

#ifdef HAVE_STRING_H
# include <string.h>
#endif /*HAVE_STRING_H*/

#ifdef HAVE_STDLIB_H
# include <stdlib.h>
#endif /*HAVE_STDLIB_H*/

#ifdef HAVE_LIMITS_H
# include <limits.h>
#endif /*HAVE_LIMITS_H*/

#ifdef RTTB_HDRSTOP
# pragma hdrstop
#endif /*RTTB_HDRSTOP*/

#include "rttb.h"
#include "rttb_int.h"
#include "rttb_ll.h"

/* Local types */
typedef unsigned int BOOL;
#define TRUE (1U)
#define FALSE (!TRUE)
/*-------------------------------------------------------------------------*/
static RTTB_ERROR _i_ContainedOpen(const char* RTTB, const char* Filename, const char* Mode, RTTB_HANDLE* PGetHandle);
static RTTB_ERROR _i_LooseOpen(const char* RTTB, const char* Filename, const char* Mode, RTTB_HANDLE* PGetHandle);
static RTTB_ERROR i_IsReadonlyMode(const char* Mode); /* Can only return RTTB_ERROR_READONLY, if readonly modes are used */
/*-------------------------------------------------------------------------*/
RTTB_HANDLE rttb_fopen(const char* ProjectedRTTB, const char* FileName, const char* Mode, const int LooseFirst, RTTB_ERROR* Perr)
{
  RTTB_HANDLE handle;
  RTTB_ERROR terr; /* Temporary error on stack */

  /* Check params, we have to anyway but we can use assert() for debug builds */
  assert(ProjectedRTTB);
  assert(FileName);
  assert(Mode);

#ifdef NDEBUG
  if ( !ProjectedRTTB || !FileName || !Mode ) {
    if (Perr) *Perr = RTTB_ERROR_INVPARAMS;
    return RTTB_NULLHANDLE;
  }
#endif /*NDEBUG*/

  terr = i_IsReadonlyMode(Mode);
  if ( terr ) {
    if ( Perr ) *Perr = terr;
    return RTTB_NULLHANDLE; /* Only readonly modes allowed */
  }

  /* Register the RTTB container if neccersary */
  if ( !rttb_core()->IsContainerRegistered(ProjectedRTTB) ) {
    if ( !rttb_core()->RegisterContainer(ProjectedRTTB) ) {
      /* Due to a memory allocation problem in the core linked list, I can't
      add this container, damn and blast */

      if (Perr) *Perr = RTTB_ERROR_MALLOC;
      return RTTB_NULLHANDLE;
    }
  }

  if ( LooseFirst ) {
    DBGMSG("opening loose first")
    terr = _i_LooseOpen(ProjectedRTTB, FileName, Mode, &handle); /* Open the file on disk */
    if ( terr != RTTB_ERROR_NOERROR ) { /* Failed open, use rttb instead */
      DBGMSG("That failed, container next")
      terr = _i_ContainedOpen(ProjectedRTTB, FileName, Mode, &handle);
      /* Doesn't appear to be in RTTB file, in that case file does not exist */
      if ( terr != RTTB_ERROR_NOERROR ) { /* An error */
        if (Perr) *Perr = RTTB_ERROR_CANTOPEN; /* Record for caller */
        return RTTB_NULLHANDLE;
      }
      return RTTB_NULLHANDLE;
    }
  }
  else { /* Contained in preference, my personal preference */
    DBGMSG("opening container first")
    terr = _i_ContainedOpen(ProjectedRTTB, FileName, Mode, &handle);
    if ( terr == RTTB_ERROR_NOERROR )
      return handle;
      
    /* Failed to open, not in grouped rttb so check for loose file instead */
    DBGMSG("that failing, going loose")
    terr = _i_LooseOpen(ProjectedRTTB, FileName, Mode, &handle); /* Open the file on disk */
    if ( terr != RTTB_ERROR_NOERROR ) { /* That failed too, give up */
      if (Perr) *Perr = terr;
      return handle; /* (null handle) */
    }
  }
  /* Give data back to caller, success */
  if ( Perr ) *Perr = RTTB_ERROR_NOERROR;
  assert(handle); /* Must never pass pack a NULL handle with no error */
  return handle;
}
/*-------------------------------------------------------------------------*/
static RTTB_ERROR _i_ContainedOpen(const char* RTTB, const char* Filename, const char* Mode, RTTB_HANDLE* PGetHandle)
{
  RTTB_ERROR tmperr;
  RTTB_HANDLE handle;

  if ( PGetHandle ) *PGetHandle = RTTB_NULLHANDLE; /* Init caller's copy of handle */
  if ( !rttb_core()->IsContainerPrepared(RTTB) ) {
    tmperr = rttb_core()->PrepareContainer(RTTB);
    if ( tmperr != RTTB_ERROR_NOERROR )
      return tmperr;
  }
  /* Container prepared */
  tmperr = rttb_core()->PrepareFile(RTTB, Filename);
  if ( tmperr ) return tmperr; /* Get out here incase of error */

  /* This is all succssful, the rest is implicit, next time caller reads
     and writes they will used the core-assigned handle.  Note that the core.c
     file does not know what valid modes are, it expects fopen.c to be a filter
     it will only store the modes for modules who request it.
  */
  handle = rttb_core()->GetHandle(RTTB, Filename);
  assert(handle); /* Souldn't happen, we know details are correct! */
  rttb_core()->SetFileMode(handle, Mode);
  if ( PGetHandle ) *PGetHandle = handle;
  return RTTB_ERROR_NOERROR;
}
/*-------------------------------------------------------------------------*/
static RTTB_ERROR _i_LooseOpen(const char* RTTB, const char* Filename, const char* Mode, RTTB_HANDLE* PGetHandle)
{
  RTTB_HANDLE genHandle;
  BOUNDFILE boundfile;
  FILE* newpfs;
  unsigned int errornumber;

  if ( PGetHandle ) *PGetHandle = RTTB_NULLHANDLE; /* Init caller's copy of handle */
  newpfs = rttb_open(Filename, Mode);
  if ( !newpfs )
    return RTTB_ERROR_CANTOPEN;

  genHandle = rttb_core()->GenerateHandle();
  if ( genHandle == RTTB_NULLHANDLE ) {
    rttb_close(newpfs);
    return RTTB_ERROR_HANDLEGEN;
  }
  rttb_core()->InitBoundFile(&boundfile);
  boundfile.filename = rttb_malloc( strlen(Filename) +1 );
  if ( !boundfile.filename ) {
    rttb_close(newpfs);
    return RTTB_NULLHANDLE;
  }
  strcpy(boundfile.filename, Filename);
  boundfile.pfs = newpfs;
  errornumber = rttb_core()->AddHandle(genHandle, RTTB, &boundfile);
  if ( errornumber == 5 ) /* No more entries for this container */
    return RTTB_ERROR_OPENLIMIT;

  assert( !errornumber ); /* This better be zero or else I want to know about it */

  rttb_core()->SetFileMode(genHandle, Mode);
  if ( PGetHandle ) *PGetHandle = genHandle;
  return RTTB_ERROR_NOERROR;
}
/*-------------------------------------------------------------------------*/
static RTTB_ERROR i_IsReadonlyMode(const char* Mode)
{
  const char acc[] = {'r','+','b','t'};
  size_t modeLen, modeIndex;

  if ( !Mode ) return FALSE;
  if ( Mode[0] == '\0' ) return FALSE;

  modeLen = strlen(Mode);
  for ( modeIndex = 0U; modeIndex < modeLen; modeIndex++ ) {
    size_t accIndex;
    BOOL matched = FALSE;

    for ( accIndex = 0U; accIndex < sizeof(acc)/sizeof(acc[0]); accIndex++ ) {
      if ( Mode[modeIndex] == acc[accIndex] ) {
        matched = TRUE;
        break; /* Slight optimization I added: 15th Jan 2001 */
      }
    }
    if ( !matched )
      return RTTB_ERROR_READONLY;
  }
  return RTTB_ERROR_NOERROR;
}
/*-------------------------------------------------------------------------*/
