/*
Daybo Logic Run-Time TarBall Library And Tools
Copyright (c) 2000-2006, David Duncan Ross Palmer, Daybo Logic
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
      
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
      
    * Neither the name of the Daybo Logic nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

#define RTTB_SOURCE
#define RTTB_USING_TYPES

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif /*HAVE_CONFIG_H*/

#include <assert.h>
#include <stdio.h>

#ifdef HAVE_STDDEF_H
# include <stddef.h>
#endif /*HAVE_STDDEF_H*/

#ifdef HAVE_STRING_H
# include <string.h>
#endif /*HAVE_STRING_H*/

#ifdef HAVE_STDLIB_H
# include <stdlib.h>
#endif /*HAVE_STDLIB_H*/

#ifdef HAVE_LIMITS_H
# include <limits.h>
#endif /*HAVE_LIMITS_H*/

#ifdef HAVE_STDBOOL_H
# include <stdbool.h>
#else
# include "dlstdinc_stdbool.h"
#endif /*HAVE_STDBOOL_H*/

#ifdef RTTB_HDRSTOP
# pragma hdrstop
#endif /*RTTB_HDRSTOP*/

#include "rttb.h"
#include "rttb_int.h"
#include "rttb_ll.h"
/*-------------------------------------------------------------------------*/
static bool _i_ValidWhence(const int Whence);
static long int _i_NormalizeOffset(const RTTB_HANDLE Handle, const long int Offset, const int Whence); /* Makes the offset relative from the file beginning (SEEK_SET) */
/* Although it may seem like fsp is a redundant piece of data in the following
calls, it is not because fsp is needed, takes longer to learn from Handle,
the caller have both and we need both. */
static RTTB_ERROR _i_seek(RTTB_HANDLE Handle, const long int Offset);
static void _i_recover(RTTB_HANDLE Handle, FILE* HintFSP); /* Stream position recovery, note: fsp is optional although if available it's appreciated for speed */
/*-------------------------------------------------------------------------*/
RTTB_ERROR rttb_fseek(RTTB_HANDLE Handle, long int Offset, int Whence)
{
  RTTB_ERROR err;

  if ( !_i_ValidWhence(Whence) )
    return RTTB_ERROR_INVPARAMS;
  if ( !rttb_core()->IsValidHandle(Handle) )
    return RTTB_ERROR_INVHANDLE;

  if ( Whence != SEEK_SET ) /* Optimization to avoid unneccersary call */
    Offset = _i_NormalizeOffset(Handle, Offset, Whence);
  /* Note now Whence is effectively SEEK_SET regardless after normalization */

  err = _i_seek(Handle, Offset);
  if ( err != RTTB_ERROR_NOERROR )
    return err;

  return RTTB_ERROR_NOERROR; /* Success */
}
/*-------------------------------------------------------------------------*/
static bool _i_ValidWhence(const int Whence)
{
  if ( (Whence != SEEK_CUR) && (Whence != SEEK_SET) && (Whence != SEEK_END) ) {
    DBGMSG("Whence should be SEEK_SET, SEEK_CUR or SEEK_END")
    return false;
  }
  return true;
}
/*-------------------------------------------------------------------------*/
static long int _i_NormalizeOffset(const RTTB_HANDLE Handle, const long int Offset, const int Whence)
{
  long int an_offset;

  switch ( Whence ) {
    case SEEK_CUR : { /* Offset is relative to current position */
      rttb_core()->GetFileOffset(Handle, &an_offset);
      an_offset += Offset;
      break;
    }
    case SEEK_END : { /* Offset is relative to end (must be negative!) */
      rttb_core()->GetFileLength(Handle, &an_offset);
      an_offset += Offset;
      break;
    }
    default : { /* Assume SEEK_SET */
      an_offset = Offset;
    }
  }
  return an_offset;
}
/*-------------------------------------------------------------------------*/
static RTTB_ERROR _i_seek(RTTB_HANDLE Handle, const long int Offset)
{
  long int li; /* Used twice, so no meaningful name! */
  RTTB_ERROR err;
  FILE* fsp;

  if ( Offset < 0 ) /* We know negatives aren't valid thanks to normalizations */
    return RTTB_ERROR_UNDERSEEK;

  err = rttb_core()->GetFileLength((const RTTB_HANDLE)Handle, &li); /* Find the current length */
  if ( err != RTTB_ERROR_NOERROR )
    return err;

  if ( Offset >= li ) /* Trying to overseek! */
    return RTTB_ERROR_OVERSEEK;

  /* Validated, do seek & update if successful, no need to seek for contained files */
  fsp = rttb_core()->GetFSP(Handle);
  if ( fsp ) { /* Loose? */
    li = fseek(fsp, Offset, SEEK_SET); /* Do the pysical seek */
    /* Return from fseek is unreliable, works/does not work on different platforms
    and sets errno, shock horror it certainly doesn't return whether we hit the
    EOF or anything on Win32 or UNIX.  I will handle it it in the only way I can
    and verify the position myself */
    if ( li ) { /* Nonzero error */
      _i_recover(Handle, fsp);
      return RTTB_ERROR_UNEXPECTED;
    }
    if ( ftell(fsp) != Offset ) {
      _i_recover(Handle, fsp);
      return RTTB_ERROR_UNEXPECTEDEOF;
    }
  }
  /* Seek OK, now update core - no need to check for errors, we know the
  handle is valid etc, checking for errors would mean having to rewind() the
  physical file again and the like, it's not going to happen anyway, just
  read the code in the function */
  rttb_core()->SetFileOffset(Handle, Offset);
  return RTTB_ERROR_NOERROR;
}
/*-------------------------------------------------------------------------*/
static void _i_recover(RTTB_HANDLE Handle, FILE* HintFSP)
{
  long int pos; /* Position stream is _supposed_ to be */
  FILE* fsp; /* Actual fsp hint or no hint */

  /* If hint passed use it, else sort ourselves, aiii */
  if ( HintFSP )
    fsp = HintFSP;
  else
    fsp = rttb_core()->GetFSP((const RTTB_HANDLE)Handle);
  assert(HintFSP); /* Must not be! */

  if ( rttb_core()->GetFileOffset(Handle, &pos) != RTTB_ERROR_NOERROR )
    return;

  rewind(fsp);
  fseek(fsp, pos, SEEK_SET);
  return;
}
/*-------------------------------------------------------------------------*/
