/*
Daybo Logic Run-Time TarBall Library And Tools
Copyright (c) 2000-2006, David Duncan Ross Palmer, Daybo Logic
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
      
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
      
    * Neither the name of the Daybo Logic nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

#define RTTB_SOURCE

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif /*HAVE_CONFIG_H*/

#include <assert.h>

#ifdef HAVE_STDLIB_H
# include <stdlib.h>
#endif /*HAVE_STDLIB_H*/

#ifdef HAVE_STRING_H
# include <string.h>
#endif /*HAVE_STRING_H*/

#include <stdio.h>

#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif /*HAVE_LIMITS_H*/

#ifdef RTTB_HDRSTOP
# pragma hdrstop
#endif /*RTTB_HDRSTOP*/

#include "rttb.h"
#include "rttb_int.h"
#include "rttb_ll.h"
/*-------------------------------------------------------------------*/
/* i_ = Internal use l_ = Linked use (for outsiders) */
static int l_Add(struct link_ptrs* Obj, const void* Node);
static int l_Del(struct link_ptrs* Obj, const void* Node);
static int l_DelByPos(struct link_ptrs* Obj, const unsigned int Pos);
static unsigned int l_Count(const struct link_ptrs* Obj);
static void l_DelAll(struct link_ptrs* Obj);
static int l_Find(struct link_ptrs* Obj, const void* Node, unsigned int* PPos);
static void l_Set(struct link_ptrs* Obj, const unsigned int Pos, const void* Node);
static void l_Get(const struct link_ptrs* Obj, const unsigned int Pos, void* Node);
static size_t l_NodeSize(const struct link_ptrs* Obj);
static struct link_ptrs_node* i_findend(struct link_ptrs* Obj);
static struct link_ptrs_node* i_findnode(struct link_ptrs* Obj, const void* Node, unsigned int* PPos);
static struct link_ptrs_node* i_newnode(struct link_ptrs* Obj);
static struct link_ptrs_node* i_nodefrompos(const struct link_ptrs* Obj, unsigned int Pos);
#define i_killnode(node) \
  if ((node)) rttb_free((node)->ptr); \
  rttb_free((node));
#define i_initnode(node) \
  (node)->ptr = NULL; \
  (node)->next = NULL;
/*-------------------------------------------------------------------------*/
void rttb_link_ptrs_init(struct link_ptrs* Obj, const size_t NodeSize)
{
  assert(Obj);
  assert(NodeSize); /* May not be zero */

  Obj->base = NULL;
  Obj->nodeSize = NodeSize;
  Obj->count = 0U;
  Obj->Add = l_Add;
  Obj->Del = l_Del;
  Obj->DelByPos = l_DelByPos;
  Obj->Count = l_Count;
  Obj->DelAll = l_DelAll;
  Obj->Find = l_Find;
  Obj->Set = l_Set;
  Obj->Get = l_Get;
  Obj->NodeSize = l_NodeSize;
}
/*-------------------------------------------------------------------------*/
static int l_Add(struct link_ptrs* Obj, const void* Node)
{
  struct link_ptrs_node* endNode;
  assert(Obj);

  endNode = i_findend(Obj);
  if (!endNode) {
    Obj->base = i_newnode(Obj);
    endNode = Obj->base;
  }
  else {
    endNode->next = i_newnode(Obj);
    endNode = endNode->next; /* Correct what's the new end node */
  }
  if (!endNode) return 0; /* Failed to create new node */

  Obj->count++; /* Keep track of added nodes */
  memcpy(endNode->ptr, Node, Obj->nodeSize); /* Copy data off caller */
  return 1; /* Success */
}
/*-------------------------------------------------------------------------*/
static int l_Del(struct link_ptrs* Obj, const void* Node)
{
  struct link_ptrs_node* tnode; /* Target node, node to kill! */
  unsigned int i; /* Holds index of node above one being removed */
  assert(Obj);
  assert(Node);

  tnode = i_findnode(Obj, Node, &i);
  if (tnode) {
    if (i) { /* Is there a node before us? */
      struct link_ptrs_node* prevNode = i_nodefrompos(Obj, i-1);
      assert(prevNode); /* This should never occur! */
      prevNode->next = tnode->next;
    }
    /* Links corrected, kill node */
    i_killnode(tnode);
    Obj->count--;
    return 1; /* Success */
  }
  return 0; /* Failed!  Not not present */
}
/*-------------------------------------------------------------------------*/
static int l_DelByPos(struct link_ptrs* Obj, const unsigned int Pos)
{
  struct link_ptrs_node* tnode;
  unsigned int i = 0U;
  assert(Obj);

  tnode = Obj->base;
  while ( tnode ) {
    if ( i == Pos ) { /* Node to kill? */
      if (i) { /* There's a previous node before us? */
        struct link_ptrs_node* prevNode = i_nodefrompos(Obj, i-1);
        assert(prevNode); /* Never happen I hope */
        prevNode->next = tnode->next;
      }
      else { /* No previous node */
        Obj->base = tnode->next; /* Bug fix: Relinking of base */
      }
      i_killnode(tnode);
      Obj->count--;
      return 1; /* Deleted successfully */
    }
    else { /* Not the node to kill */
      tnode = tnode->next; /* Transcend */
    }
    i++;
  }
  return 0; /* Invalid pos */
}
/*-------------------------------------------------------------------------*/
static unsigned int l_Count(const struct link_ptrs* Obj)
{
  assert(Obj);
  return Obj->count;
}
/*-------------------------------------------------------------------------*/
static void l_DelAll(struct link_ptrs* Obj)
{
  assert(Obj);

  while (Obj->count)
    Obj->DelByPos(Obj, 0U);

  return;
}
/*-------------------------------------------------------------------------*/
static int l_Find(struct link_ptrs* Obj, const void* Node, unsigned int* PPos)
{
  struct link_ptrs_node* foundNode;
  assert(Obj);

  foundNode = i_findnode(Obj, Node, PPos);
  if (foundNode) return 1;
  return 0;
}
/*-------------------------------------------------------------------------*/
static void l_Set(struct link_ptrs* Obj, const unsigned int Pos, const void* Node)
{
  struct link_ptrs_node* node;
  assert(Obj);
  /*assert( Pos >= Obj->count ); Range check - flippin' Borland is optimizing my code and making this test mess up */
  assert(Node);

  node = i_nodefrompos(Obj, Pos);
  if (node) memcpy(node->ptr, Node, Obj->nodeSize);
}
/*-------------------------------------------------------------------------*/
static void l_Get(const struct link_ptrs* Obj, const unsigned int Pos, void* Node)
{
  struct link_ptrs_node* node;
  assert(Obj);
#if 0
  /* This assert doesn't work because of some freak optimization by Borland */
  assert( Pos >= Obj->count ); /* Range check */
#endif /*0*/
  assert(Node);

  node = i_nodefrompos(Obj, Pos);
  if (node) memcpy(Node, node->ptr, Obj->nodeSize);
}
/*-------------------------------------------------------------------------*/
static size_t l_NodeSize(const struct link_ptrs* Obj)
{
  /* I know this is simple but it's a matter of principle, we're entitled
  to access the "private" member whereas callers are not */
  assert(Obj);
  return Obj->nodeSize;
}
/*-------------------------------------------------------------------------*/
static struct link_ptrs_node* i_findend(struct link_ptrs* Obj)
{
  struct link_ptrs_node* currNode;
  assert(Obj);

  if ( !Obj->base ) return NULL; /* No linked list */
  currNode = Obj->base;
  while ( currNode->next )
    currNode = currNode->next;

  return currNode;
}
/*-------------------------------------------------------------------------*/
static struct link_ptrs_node* i_findnode(struct link_ptrs* Obj, const void* Node, unsigned int* PPos)
{
  struct link_ptrs_node* currNode;
  unsigned int i = 0U;
  assert(Obj);
  assert(Node);

  if ( !Obj->base) return NULL; /* No linked list */
  currNode = Obj->base;
  while ( currNode ) {
    assert(currNode->ptr);
    if (memcmp(currNode->ptr, Node, Obj->nodeSize) == 0) {
      if (PPos) *PPos = i;
      break; /* Found node */
    }
    currNode = currNode->next;
    i++;
  }
  return currNode;
}
/*-------------------------------------------------------------------------*/
static struct link_ptrs_node* i_newnode(struct link_ptrs* Obj)
{
  struct link_ptrs_node* newnode = (struct link_ptrs_node*)rttb_malloc(sizeof(struct link_ptrs_node));
  if (newnode) {
    i_initnode(newnode);
    newnode->ptr = rttb_malloc( Obj->nodeSize );
    if (!newnode->ptr) {
      rttb_free(newnode);
      return NULL;
    }
    return newnode;
  }
  return NULL;
}
/*-------------------------------------------------------------------------*/
static struct link_ptrs_node* i_nodefrompos(const struct link_ptrs* Obj, unsigned int Pos)
{
  unsigned int i;
  struct link_ptrs_node* currnode = Obj->base;

  for ( i = 0U; i < Pos; i++ ) {
    if (!currnode) return NULL; /* Whoops, read past range, caller is evil! */
    currnode = currnode->next; /* Crawl */
  }
  return currnode;
}
/*-------------------------------------------------------------------------*/
