/*
Daybo Logic Run-Time TarBall Library And Tools
Copyright (c) 2000-2006, David Duncan Ross Palmer, Daybo Logic
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
      
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
      
    * Neither the name of the Daybo Logic nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

/*
  rttb lib's addon library set for dealing with double zero terminated
  strings, these functions are not used by users of the library, only by
  other modules.  If anybody requests to use these functions for themselves
  later on, I will make them available elsewhere, say in a string library.
  Library dependencies and symbol conflicts will become far too complicated
  if I give these functions (here) common names which other libraries
  might try to link in too.  I generally don't make libraries rely on other
  libraries, it just gets mental quickly.
*/

#define RTTB_SOURCE

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif /*HAVE_CONFIG_H*/

#include <assert.h>

#ifdef HAVE_STRING_H
# include <string.h>
#endif /*HAVE_STRING_H*/

#include <stdio.h>

#ifdef HAVE_STDLIB_H
# include <stdlib.h>
#endif /*HAVE_STDLIB_H*/

#ifdef HAVE_LIMITS_H
# include <limits.h>
#endif /*HAVE_LIMITS_H*/

#ifdef RTTB_HDRSTOP
# pragma hdrstop
#endif /*RTTB_HDRSTOP*/

#include "rttb.h"
#include "rttb_int.h"
#include "rttb_ll.h"
#include "rttb_strzz.h"

#define FINDEND(str) \
        (char*)( (str) + rttb_internal_strzzFullLength((str)) )
/*--------------------------------------------------------------------------*/
/* Internal functions */
static void _defhandler(struct strzz*); /* My default handler */
static void _cleanLL(struct link_ptrs* pll); /* pll must be specified, link list of dynamic strings cleaned up */

/* Internal data */
static void (*_handler)(struct strzz*) = _defhandler; /* Active handler, mine by default */

/* Effective member functions */
static void _Add(struct strzz*, const char* NewStr);
static void _Remove(struct strzz*, const int Index);
static unsigned int _Count(const struct strzz*);
static int _Find(const struct strzz*, const char* CompareString);
static const char* _GetByIndex(const struct strzz*, const int Index);
static void _SetByIndex(struct strzz*, const int Index, const char* NewString);
static unsigned int _FullLength(const struct strzz*);
static const char* _GetCString(struct strzz*); /* Returns the NULL terminated string */
/*--------------------------------------------------------------------------*/
static void _defhandler(struct strzz* StrZZ)
{
  if ( StrZZ ) StrZZ->Destruct(StrZZ); /* Perform any cleanup */
  fprintf(RTTB_DEV_ERROR, "%s : Out of memory (strzz object %p), unhandled\n", __FILE__, StrZZ);
  abort();
}
/*--------------------------------------------------------------------------*/
static void _cleanLL(struct link_ptrs* pll)
{
  while ( pll->Count(pll) ) {
    char* ptr;
    pll->Get(pll, 0U, &ptr);
    if (ptr)
      rttb_free(ptr);
    pll->DelByPos(pll, 0U);
  }
  return;
}
/*--------------------------------------------------------------------------*/
void rttb_internal_strzzConstruct(struct strzz* StrZZ)
{
  if ( StrZZ ) {
    /* Init data */
    rttb_link_ptrs_init(&StrZZ->_ll, sizeof(char*));
    StrZZ->_pCreatedString = NULL;
    StrZZ->_cached = 0U; /* Not cached */
    /* Init member functions */
    StrZZ->Add = _Add;
    StrZZ->Remove = _Remove;
    StrZZ->Count = _Count;
    StrZZ->Find = _Find;
    StrZZ->GetByIndex = _GetByIndex;
    StrZZ->SetByIndex = _SetByIndex;
    StrZZ->FullLength = _FullLength;
    StrZZ->Destruct = rttb_internal_strzzDestruct;
    StrZZ->GetCString = _GetCString;
  }
  return;
}
/*--------------------------------------------------------------------------*/
void rttb_internal_strzzDestruct(struct strzz* StrZZ)
{
  if ( StrZZ ) {
    _cleanLL(&StrZZ->_ll);
    if ( StrZZ->_pCreatedString ) {
      rttb_free(StrZZ->_pCreatedString);
      StrZZ->_pCreatedString = NULL;
    }
    StrZZ->_cached = 0U;
    /* So that everything is clean and traceable if the user tries to access the functions...
    I'll strip these later if I'm happy */
    StrZZ->Add = NULL;
    StrZZ->Remove = NULL;
    StrZZ->Count = NULL;
    StrZZ->Find = NULL;
    StrZZ->GetByIndex = NULL;
    StrZZ->SetByIndex = NULL;
    StrZZ->FullLength = NULL;
    StrZZ->Destruct = NULL;
    StrZZ->GetCString = NULL;
  }
  return;
}
/*--------------------------------------------------------------------------*/
static void _Add(struct strzz* StrZZ, const char* NewStr)
{
  if ( StrZZ && NewStr ) {
    char* newptr;

    newptr = rttb_malloc( sizeof(char) * (strlen(NewStr) +1) );
    if ( !newptr ) { /* Allocation failure */
      _handler(StrZZ);
      return;
    }
    strcpy(newptr, NewStr); /* Bug fix: 21st April 2001: DDRP */
    if ( !StrZZ->_ll.Add(&StrZZ->_ll, &newptr) ) {
      /* Damn, can't add to linked list */
      rttb_free(newptr);
      _handler(StrZZ);
      return;
    }
    StrZZ->_cached = 0U; /* Invalidate cache */
    /* Sorted, governor */
  }
  return;
}
/*--------------------------------------------------------------------------*/
static void _Remove(struct strzz* StrZZ, const int Index)
{
  if ( StrZZ ) {
    unsigned int c;

    assert(Index != -1);
    c = StrZZ->_ll.Count(&StrZZ->_ll);
    if ( (unsigned int)Index >= c ) return; /* Read out of range */
    StrZZ->_ll.DelByPos(&StrZZ->_ll, Index);
    StrZZ->_cached = 0U; /* Invalidate cache */
  }
}
/*--------------------------------------------------------------------------*/
static unsigned int _Count(const struct strzz* StrZZ)
{
  if ( StrZZ ) return StrZZ->_ll.Count(&StrZZ->_ll);
  return 0U;
}
/*--------------------------------------------------------------------------*/
static int _Find(const struct strzz* StrZZ, const char* CompareString)
{
  if ( StrZZ && CompareString ) {
    unsigned int i, c;

    c = StrZZ->_ll.Count(&StrZZ->_ll);
    for ( i = 0U; i < c; i++ ) {
      char* tmpptr;

      StrZZ->_ll.Get(&StrZZ->_ll, i, &tmpptr); /* Read back pointer */
      if ( tmpptr ) {
        if ( !strcmp(tmpptr, CompareString) )
          return (int)i;
      }
    }
  }
  return -1; /* Not found / invalid args */
}
/*--------------------------------------------------------------------------*/
static const char* _GetByIndex(const struct strzz* StrZZ, const int Index)
{
  char* tmpptr = NULL;

  if ( StrZZ ) {
    assert( Index != -1 );
    if ( (unsigned int)Index >= StrZZ->_ll.Count(&StrZZ->_ll) ) return tmpptr; /* Out of range */
    if ( StrZZ )
      StrZZ->_ll.Get(&StrZZ->_ll, (unsigned int)Index, &tmpptr);
  }
  return tmpptr;
}
/*--------------------------------------------------------------------------*/
static void _SetByIndex(struct strzz* StrZZ, const int Index, const char* NewString)
{
  assert( Index != -1 );
  if ( StrZZ && NewString) {
    char *tmpptr, *newptr;
    if ( (unsigned int)Index >= StrZZ->_ll.Count(&StrZZ->_ll) ) return; /* Out of range */

    StrZZ->_ll.Get(&StrZZ->_ll, (unsigned int)Index, &tmpptr);
    newptr = rttb_malloc( strlen(NewString) +1 );
    if ( newptr )
      strcpy(newptr, NewString);
    else {
      _handler(StrZZ);
      return;
    }

    if ( tmpptr ) /* Previous string to destroy */
      rttb_free(tmpptr);

    StrZZ->_ll.Set(&StrZZ->_ll, (unsigned int)Index, &newptr);
    StrZZ->_cached = 0U; /* Invalidate cache */
  }
}
/*--------------------------------------------------------------------------*/
static unsigned int _FullLength(const struct strzz* StrZZ)
{
  unsigned int length = 0U;

  if ( StrZZ ) {
    unsigned int c, i;

    c = StrZZ->_ll.Count(&StrZZ->_ll);
    for ( i = 0U; i < c; i++ ) {
      char* stringptr;

      StrZZ->_ll.Get(&StrZZ->_ll, i, &stringptr);
      if ( stringptr )
        length += strlen(stringptr);
    }
  }
  return length;
}
/*--------------------------------------------------------------------------*/
void rttb_internal_strzzSetAbortHandler(void (*NewHandler)(struct strzz*))
{
  _handler = (NewHandler) ? (NewHandler) : (_defhandler);
}
/*--------------------------------------------------------------------------*/
static const char* _GetCString(struct strzz* StrZZ)
{
  /* Calculate size of string to copy to caller, there must be NULL
  terminators for every string + 1 for double NULL on the end too. */
  if ( StrZZ ) {
    unsigned int c, i;
    size_t allocSize = (size_t)0U;
    size_t pos = 0U; /* Offsetter */

    /* Use cached copy of possible */
    if ( !StrZZ->_pCreatedString )
      StrZZ->_cached = 0U;
    if ( StrZZ->_cached )
      return StrZZ->_pCreatedString;

    c = StrZZ->_ll.Count(&StrZZ->_ll);
    for ( i = 0U; i < c; i++ ) {
      char* strptr;

      StrZZ->_ll.Get(&StrZZ->_ll, i, &strptr);
      if ( strptr ) {
        allocSize += strlen(strptr);
        allocSize++; /* For NULL terminator */
        if ( i == c - 1 )
          allocSize++; /* Last double NULL terminator */
      }
    }
    /* Now I know the alloc size, check for previous and replace */
    if ( StrZZ->_pCreatedString ) {
      rttb_free(StrZZ->_pCreatedString);
      StrZZ->_pCreatedString = NULL;
    }
    StrZZ->_pCreatedString = rttb_malloc(allocSize);
    if ( !StrZZ->_pCreatedString ) { /* Allocation failure, abort! */
      _handler(StrZZ);
      return NULL;
    }
    for ( i = 0U; i < c; i++ ) {
      char* strptr;

      StrZZ->_ll.Get(&StrZZ->_ll, i, &strptr);
      if ( strptr ) {
        size_t mystrsize = strlen(strptr)+1;
        strcpy(StrZZ->_pCreatedString + pos, strptr);
        pos += mystrsize;
      }
      if ( i == c - 1 ) /* Last one */
        StrZZ->_pCreatedString[++pos] = '\0'; /* Double NULL terminator */
    }
    StrZZ->_cached = 1U; /* Yay, cached, don't bother regenerating until it's changed */
  }
  return StrZZ->_pCreatedString;
}
/*--------------------------------------------------------------------------*/
