/*
Daybo Logic Run-Time TarBall Library And Tools
Copyright (c) 2000-2006, David Duncan Ross Palmer, Daybo Logic
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
      
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
      
    * Neither the name of the Daybo Logic nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

/*
  RTTB: Daybo Logic's RunTime TarBall file format, creation program
  by David Duncan Ross Palmer.
*/

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif /*HAVE_CONFIG_H*/

#include <stdio.h>

#ifdef HAVE_STDLIB_H
# include <stdlib.h>
#endif /*HAVE_STDLIB_H*/

#ifdef HAVE_STRING_H
# include <string.h>
#endif /*HAVE_STRING_H*/

#ifdef HAVE_LIMITS_H
# include <limits.h>
#endif /*HAVE_LIMITS_H*/

#ifdef RTTB_HDRSTOP
# pragma hdrstop
#endif /*RTTB_HDRSTOP*/

#include "ddrptype.h" /* Types used by DDRP in the game */
#include "simplemm.h" /* Very simple leak tracking, extreme cut down of DPCRTLMM one might say */
#include "inputll.h" /* Linked list of source (input to RTTB) files */
#include "rttbutil.h" /* Private dlutils partial copy */
#include "rttb.h" /* The library */
#include "rttb_ver.h" /* Steal year information from library */

/* Macros and stuff I can change with no hassle */
#define VERSION_MAJOR (0)
#define VERSION_MINOR (2)
#define VERSION_BUILD (0)

/* Macros I am using for internal data */
#define ERROR_DDRPMALLOC (1) /* Indicates out of memory error */
#define ERROR_FOPEN (2) /* Cannot open the specified file */
#define ERROR_RTTBNULL (3) /* An internal error */
#define ERROR_WRITEERROR (4) /* Probally out of disk space */
#define ERROR_READDEVICEFAIL (5) /* Read error while we know the length (device error) */

/* Macros for debugging/release */
#ifndef NDEBUG /* Debug */
# define FWRITE(p, s, n, pfs) \
         _debug_fwrite((p), (s), (n), (pfs)) /* Still compatible with if statements, this function flushes as well as writing so I can hex dump after each write */
#else /* Release */
# define FWRITE(p, s, n, pfs) \
         fwrite((p), (s), (n), (pfs))
#endif /*!NDEBUG*/

static void Banner(void);
static void Syntax(const char* ExeName);
static const char* RttbName;
static BOOL ProcessArguments(int argc, char** argv);
static unsigned short AddSource(char* const Filename, const BOOL IsDynamic); /* Add a source file to the list of files to be parsed, an error number (nonzero) might be returned */
static void PrintError(const unsigned short E, const char* FileName);
static BOOL CreateRTTB(void); /* Does main work */
static void WriteCopyright(FILE* HFile); /* Simply writes my copyright notice to the RTTB file */
static BOOL WriteRevision(FILE* HFile); /* Writes the revision number RTTB_REVISION */
static DWORD CalcHeaderLength(void); /* Calculate size of the header (used for the fast end marker) */
static BOOL WriteWORD(FILE* HFile, const WORD w); /* See WriteDWORD() */
static BOOL WriteDWORD(FILE* HFile, const DWORD dw); /* Writes a simple DWORD, big endian, bytes swaps if required */
static DWORD Checksum32(S_SOURCE* PSource); /* Opens the file temporarily, generates the checksum for the data (32-bit) and returns it */
static BOOL WriteLength(FILE* HFile, const ULONG Length); /* Writes the said length of the rttb file (used in the table of lengths) */
static BOOL WriteFN(FILE* HFile, const char* SourceFileName); /* Writes a filename at the current position (or any othr text for that matter) */
static void ProcessResponse(const char* ResponseFileName); /* Processes a reponse file, line by line, adds strings to list of sources */
static size_t ChopNewline(char* LineStr, size_t Length); /* Length may be changed, used returned value */
#ifndef NDEBUG /* Debug only */
  size_t _debug_fwrite(const void* ptr, size_t size, size_t n, FILE* stream);
#endif /*!NDEBUG*/

static const char copyrightInfo[] = "David D.R. Palmer\'s run-time tarball format,\ncopyright 2000, Daybo Logic all rights reserved.";
/*--------------------------------------------------------------------------*/
int main(int argc, char* argv[])
{
  /* Locals */
  BOOL rttbCreation; /* return from CreateRTTB() */

  Banner();

  if ( !ProcessArguments(argc, argv) )
  {
    Syntax(argv[0]); /* Show user what they should have used */
    return EXIT_FAILURE;
  }

  putc('\n', stdout);
  if ( !LL_CountItems() )
  {
    printf("makerttb: Nothing to do!\n");
    return EXIT_FAILURE;
  }

  rttbCreation = CreateRTTB();
  LL_DeleteSources(); /* Delete all elements of the linked list */
  if ( !rttbCreation ) /* Does the main work */
  {
    printf("\nErrors before creation complete, deleting \'%s\'\n", RttbName);
    remove(RttbName); /* Incomplete RTTB deleted */
    return EXIT_FAILURE;
  }

#ifndef NDEBUG
  /* Debug code only */
  if ( ddrpgetallocationcount() ) {
    printf(
      "\n\nATTENTION! %lu blocks not freed!\n",
      ddrpgetallocationcount()
    );
  }
#endif /*!NDEBUG*/

  return EXIT_SUCCESS;
}
/*--------------------------------------------------------------------------*/
static void Banner()
{
  printf("%s\n", "Runtime Tarball Compiler,");
  printf("%s", "(C)Copyright 2000");
  #if ( RTTB_THISYEAR > 2000 )
  printf(" - %u", RTTB_THISYEAR);
  #endif
  printf("%s\n", ", Daybo Logic, all rights reserved.");
  printf(
    "Version: V%u.%u.%u for RTTB revision %u, built %s %s\n",
    VERSION_MAJOR, VERSION_MINOR, VERSION_BUILD, RTTB_REVISION,
    __DATE__, __TIME__
  );
  printf("Please send bug reports to: www.daybologic.co.uk/mailddrp\n\n");
  return;
}
/*--------------------------------------------------------------------------*/
static void Syntax(const char* ExeName)
{
  const char defaultExeName[] = "makerttb";
  const char* _exeName = NULL;

  if (ExeName)
  {
    if (ExeName[0])
      _exeName = (const char*)ExeName;
  }
  if ( !_exeName) /* _exeName not set by caller */
    _exeName = defaultExeName; /* Default */

  printf("%s <rttbname> <filenames ...>|@<responsefile>\n\n", _exeName);
  printf("Where rttbname is the name of the runtime tarball to create,\n");
  printf("filenames are added to it, if the OS has a limited commandline\n");
  printf("put all the filenames in a response file line by line.\n");

  printf("\n");
  return;
}
/*--------------------------------------------------------------------------*/
static BOOL ProcessArguments(int argc, char** argv)
{
  /* First check that there are at least to arguments passed, RTTB
  and at least one file or a response file */

  int i;

  if ( (argc-1) < 2 )
  {
    printf("Insufficient number of parameters\n");
    return FALSE;
  }

  for ( i = 1; i < argc; i++ )
  {
    if (argv[i][0] == '@' && argv[i][1] == '\0') /* Detected ambiguous response filename */
    {
      printf("@ - Ambiguous responsefile\n");
      return FALSE;
    }
    else if (argv[i][0] == '-') /* Detected option */
    {
      printf("All \'-\' options are reserved and non specifiable\n");
      return FALSE;
    }
    else if (argv[i][0] == '@' && i == 1) /* Responsefile not allowed at the first argument */
    {
      printf("First argument must be the rttbfile, not a response file\n");
      return FALSE;
    }
    else /* Taken as valid argument */
    {
      /*
        If it's the first, argument it's the RTTB file, else it's a filename
        or possibly a response file which is read by a separate parser
      */
      if ( i == 1 ) /* rttb file */
      {
        RttbName = argv[i];
      }
      else /* Filename or responsefile */
      {
        if ( argv[i][0] == '@' ) /* Responsefile */
        {
          ProcessResponse(argv[i]+sizeof(char)); /* Skipping first char (that's an '@') */
        }
        else
        {
          /* Add this source filename to our list of files to add */
          unsigned short adderror = AddSource(argv[i], FALSE);
          if ( adderror )
          {
            /* Error adding sourcefile to the list to parse */
            PrintError(adderror, argv[i]);
            /* Aside from that ignore it */
          }
        } /* Not response file / response file */
      } /* Not first argument */
    } /* Taken av a valid argument */
  } /* end for */

  return TRUE;
}
/*--------------------------------------------------------------------------*/
static unsigned short AddSource(char* const Filename, const BOOL IsDynamic)
{
  S_SOURCE newEntry; /* New entry for source files linked list */
  FILE* tmpOpen; /* Used to temporarilly open the file */

  newEntry.FileName = Filename;
  newEntry.FileNameDynamic = IsDynamic;

  tmpOpen = fopen(Filename, "rb"); /* Readonly binary */
  if (!tmpOpen) /* Failed to open the file? */
    return ERROR_FOPEN; /* damn */

  if (!rttbutil_GetFileLength(tmpOpen, &newEntry.Length)) /* Simple get filelength */
  {
    fclose(tmpOpen);
    return ERROR_FOPEN; /* Some other weirdo I/O problem man */
  }
  fclose(tmpOpen);
  /* Now we have the length, that's all that we need, add the item to the
     linked list */

  if ( !LL_AddSource(&newEntry) )
    return ERROR_DDRPMALLOC; /* Not enough memory to add the filename */

  return 0; /* Success */
}
/*--------------------------------------------------------------------------*/
static void PrintError(const unsigned short E, const char* FileName)
{
  if (E)
  {
    printf("makerrtb ERROR: ");
    switch (E)
    {
      case ERROR_DDRPMALLOC :
      {
        fprintf(RTTB_DEV_ERROR, "Ran out out dynamic error - %lu variable blocks allocated",
          ddrpgetallocationcount()
        );
        break;
      }
      case ERROR_FOPEN :
      {
        fprintf(RTTB_DEV_ERROR, "Cannot open file \'%s\'", ((FileName) ? (FileName) : ("(null)")));
        break;
      }
      case ERROR_RTTBNULL :
      {
        printf("RTTB unnamed at usage (internal error)");
        break;
      }
      case ERROR_WRITEERROR :
      {
        printf("Write error, while writing out %s", ((FileName) ? (FileName) : ("(null)"))); /* FileName is not a filename in this case */
        break;
      }
      case ERROR_READDEVICEFAIL :
      {
        printf("Read device failed reading data we already knew the length of, file: '%s', surface scan disk\n", ((FileName) ? (FileName) : ("(null)")));
        break;
      }
      default :
      {
        printf("Unknown error %u", E);
      }
    }
    putc((int)'\n', RTTB_DEV_ERROR);
  }
  return;
}
/*--------------------------------------------------------------------------*/
static BOOL CreateRTTB()
{
  FILE* rttbHandle;
  S_SOURCE* PcurrentSource; /* Current source file (crawling linked list) */
  BOOL status = TRUE; /* Success by default */

  printf("Creating/overwriting RTTB file \'%s\'\n", ((RttbName) ? (RttbName) : ("(null)")));
  if ( !RttbName )
  {
    PrintError(ERROR_RTTBNULL, NULL);
    return FALSE;
  }
  rttbHandle = fopen(RttbName, "wb");
  if ( !rttbHandle )
  {
    PrintError(ERROR_FOPEN, RttbName);
    return FALSE;
  }
  printf("RTTB created.\n");

  /* Write header ----------------------------------------------- */
  WriteCopyright(rttbHandle);
  if ( !WriteRevision(rttbHandle) ) /* Write out the revision number */
  {
    status = FALSE; /* Failure */
    PrintError(ERROR_WRITEERROR, "revision");
    goto giveStatus;
  }
  if ( !WriteDWORD(rttbHandle, LL_CountItems()) ) /* Write N, the most important value in the header */
  {
    status = FALSE; /* Failure */
    PrintError(ERROR_WRITEERROR, "file count (N)");
    goto giveStatus;
  }
  /* Pre-calculate the size of the whole header for fast end indicator */
  if ( !WriteDWORD(rttbHandle, CalcHeaderLength()) )
  {
    status = FALSE; /* Failure */
    PrintError(ERROR_WRITEERROR, "Fast header end marker");
    goto giveStatus;
  }
  /* Write file lengths */
  PcurrentSource = LL_Address(); /* Linked list base node */
  while ( PcurrentSource )
  {
    if ( !WriteLength(rttbHandle, PcurrentSource->Length) )
    {
      status = FALSE;
      PrintError(ERROR_WRITEERROR, "File length array");
      goto giveStatus;
    }
    PcurrentSource = PcurrentSource->PNext; /* Move onto next sourcefile name */
  }
  /* Write 32-bit checksums */
  PcurrentSource = LL_Address();
  while ( PcurrentSource )
  {
    if ( !WriteDWORD(rttbHandle, Checksum32(PcurrentSource)) )
    {
      status = FALSE;
      PrintError(ERROR_WRITEERROR, "checksum 32-bit");
      goto giveStatus;
    }
    PcurrentSource = PcurrentSource->PNext;
  }
  /* Write filename lengths */
  PcurrentSource = LL_Address();
  while ( PcurrentSource )
  {
    if ( !WriteWORD(rttbHandle, (const WORD)strlen(PcurrentSource->FileName)) ) /* Truncation of filename to 16-bit lengths is intentional and I do not repent */
    {
      status = FALSE;
      PrintError(ERROR_WRITEERROR, "filename lengths");
      goto giveStatus;
    }
    PcurrentSource = PcurrentSource->PNext;
  }
  /* Write filenames (no NULLs remember!) */
  PcurrentSource = LL_Address();
  while ( PcurrentSource )
  {
    if ( !WriteFN(rttbHandle, PcurrentSource->FileName) )
    {
      status = FALSE;
      PrintError(ERROR_WRITEERROR, "Filename data");
      goto giveStatus;
    }
    PcurrentSource = PcurrentSource->PNext;
  }

  /* Write body (raw data) -------------------------------------- */
  PcurrentSource = LL_Address(); /* linked list base node */
  while ( PcurrentSource )
  {
    FILE_BLOCK_SIZE fbs;
    unsigned long i;
    BYTE* copyBuffer;
    FILE* HsourceFile; /* Handle of the current source file */

    printf("Add \'%s\'...", PcurrentSource->FileName);
    HsourceFile = fopen(PcurrentSource->FileName, "rb");
    if ( !HsourceFile )
    {
      /* Can't open the file now! */
      status = FALSE;
      PrintError(ERROR_FOPEN, PcurrentSource->FileName);
      goto giveStatus;
    }
    if ( rttbutil_CalcFileBlockSize(PcurrentSource->Length, &fbs, (size_t)4096U) ) /* Get blocking factor for source file */
    {
      if ( fbs.BufferSize )
      {
        copyBuffer = (BYTE*)ddrpmalloc(fbs.BufferSize); /* Allocate copy buffer */
        if ( !copyBuffer )
        {
          /* Failure to allocate the copy buffer */
          status = FALSE;
          PrintError(ERROR_DDRPMALLOC, PcurrentSource->FileName);
          fclose(HsourceFile); /* Local cleanup */
          goto giveStatus;
        }
        for ( i = 0UL; i < fbs.Loops; i++ ) /* Loop to copy main portion of data */
        {
          if ( fread(copyBuffer, (size_t)fbs.BufferSize, 1, HsourceFile) < 1 ) /* Read in data (we know the length, only failure can be a freak device failure) */
          {
            /* Argh!  Freak device error */
            status = FALSE;
            PrintError(ERROR_READDEVICEFAIL, PcurrentSource->FileName);
            ddrpfree(copyBuffer);
            fclose(HsourceFile);
            goto giveStatus;
          }
          /* Got the data in the buffer, now write out to rttbfile */
          if ( FWRITE(copyBuffer, (size_t)fbs.BufferSize, 1, rttbHandle) < 1) /* Write out the data (errors are most probally out of disk space/quota */
          {
            /* Out of disk space etc. */
            status = FALSE;
            PrintError(ERROR_WRITEERROR, PcurrentSource->FileName);
            ddrpfree(copyBuffer);
            fclose(HsourceFile);
            goto giveStatus;
          }
        }
        /* Copy remainder */
        if ( fbs.Remainder )
        {
          /* Don't bother to shrink the buffer but remember, don't use it's full size */
          if ( fread(copyBuffer, (size_t)fbs.Remainder, 1, HsourceFile) < 1 )
          {
            status = FALSE;
            PrintError(ERROR_READDEVICEFAIL, PcurrentSource->FileName);
            ddrpfree(copyBuffer);
            fclose(HsourceFile);
            goto giveStatus;
          }
          if ( FWRITE(copyBuffer, (size_t)fbs.Remainder, 1, rttbHandle) < 1 )
          {
            status = FALSE;
            PrintError(ERROR_WRITEERROR, PcurrentSource->FileName);
            ddrpfree(copyBuffer);
            fclose(HsourceFile);
            goto giveStatus;
          }
        }
        ddrpfree(copyBuffer);
      }
    }
    PcurrentSource = PcurrentSource->PNext; /* Move onto next sourcefile name */
    fclose(HsourceFile);
    printf("done.\n");
  }

giveStatus:
  fclose(rttbHandle);
  printf("Closing run-time tarball \'%s\', created.\n", RttbName);
  return status;
}
/*--------------------------------------------------------------------------*/
static void WriteCopyright(FILE* HFile)
{
  fprintf(HFile, copyrightInfo);
}
/*--------------------------------------------------------------------------*/
static BOOL WriteRevision(FILE* HFile)
{
  BYTE revision = (BYTE)(RTTB_REVISION);
  if ( FWRITE(&revision, sizeof(revision), 1, HFile) < 1 )
    return FALSE;
  return TRUE;
}
/*--------------------------------------------------------------------------*/
static DWORD CalcHeaderLength()
{
  DWORD size = 0UL; /* I'll start with zero and add on as I go along */
  DWORD N;
  S_SOURCE* pnode; /* Linked list use */

  size += RTTB_COPYRIGHTLENGTH;
  size++; /* Revision byte */
  size += (sizeof(DWORD)*2); /* Two DWORDs, N and fast relocate */

  /* We sall use N to represent the number of files in the RTTB */
  N = LL_CountItems();

  /* QWORD file lengths, checksums and file name lengths per file (*N) */
  size += ( (sizeof(DWORD)*2) + sizeof(DWORD) + sizeof(WORD) ) * N;

  /* Finally loop through the linked list of filenames to add and add their
  total length to the sub */
  pnode = LL_Address();
  while ( pnode ) {
    if ( pnode->FileName )
      size += (DWORD)strlen(pnode->FileName);
    pnode = pnode->PNext; /* Transcend */
  }

  return size;
}
/*--------------------------------------------------------------------------*/
static BOOL WriteWORD(FILE* HFile, const WORD w)
{
#ifdef __L_ENDIAN__ /* Little endian machine? */
  WORD swappedw = rttb_Swap16(w); /* Do byteswap to make big endian */
#else /* Big endian machine */
  WORD swappedw = (WORD)w; /* No swap, already in desired order */
#endif /*__L_ENDIAN__*/

  if ( FWRITE(&swappedw, sizeof(swappedw), 1, HFile) < 1 )
    return FALSE;

  return TRUE;
}
/*--------------------------------------------------------------------------*/
static BOOL WriteDWORD(FILE* HFile, const DWORD dw /* to write */)
{
#ifdef __L_ENDIAN__ /* Little endian machine? */
  DWORD swappeddw = rttb_Swap32(dw); /* Do BYTE swap to make big endian */
#else /* Big endian machine */
  DWORD swappeddw = (DWORD)dw; /* No swap, already in desired order */
#endif /*__L_ENDIAN__*/

  if ( FWRITE(&swappeddw, sizeof(swappeddw), 1, HFile) < 1 )
    return FALSE;

  return TRUE;
}
/*--------------------------------------------------------------------------*/
static DWORD Checksum32(S_SOURCE* PSource)
{
  FILE* tmpHandle; /* Temporary handle of the file */
  DWORD chk32 = 0UL;

  tmpHandle = fopen(PSource->FileName, "rb"); /* Open source file as readonly bin */
  if (!tmpHandle)
    return chk32; /* 0UL */

  /* We have temporary handle, pass to utilities library */
  printf("Pre-checksumming %s, (time waste, will optimize next ver)...\n", PSource->FileName);
  chk32 = rttbutil_CalcFileChecksum32(tmpHandle);
  /* Ignoring 0, might be valid or might be error, caller must know
     this file can be opened to know if it's an error or not really */

  fclose(tmpHandle); /* Close the file once again */
  return chk32; /* Give checksum to caller */
}
/*--------------------------------------------------------------------------*/
static BOOL WriteLength(FILE* HFile, const ULONG Length)
{
  /* Length is 64-bit on Alpha and such and 32-bit on 32 or 16-bit processors,
     to make sure that the file supports both they are all stored as 64-bit
     in the actual file, when the length is 32-bit I pad the hi DWORD.
     Remember!  All values must be big endian on disk. */

  DWORD loDWORD;
#if (ULONG_MAX == 0xFFFFFFFFUL) /* 32-bit processor */
  /* Pad the high DWORD (it's a dummy hi-DWORD) */
  DWORD zero = 0UL;
  if ( FWRITE(&zero, sizeof(zero), 1, HFile) < 1 )
    return FALSE; /* Blast :( */
#endif

  /* Watch out, my swap routines only work with 32-bit numbers, (DWORDs), I don't
     have one for ULONGs yet */

  loDWORD = (DWORD)Length; /* Cast will supress range limitation warning */

#ifdef __L_ENDIAN__ /* Little endian DWORDs need to be made big endian */
  loDWORD = rttb_Swap32((const DWORD)loDWORD);
#endif /*__L_ENDIAN__*/

  if ( FWRITE(&loDWORD, sizeof(loDWORD), 1, HFile) < 1 )
    return FALSE;

  return TRUE;
}
/*--------------------------------------------------------------------------*/
static BOOL WriteFN(FILE* HFile, const char* SourceFileName)
{
  if (SourceFileName)
  {
    size_t l = strlen(SourceFileName);
    if (!l) return TRUE; /* We always succeed in writing nothing man */
    if ( FWRITE(SourceFileName, l, 1, HFile) == 1 )
      return TRUE; /* Normal success */
  }
  return FALSE;
}
/*--------------------------------------------------------------------------*/
static void ProcessResponse(const char* ResponseFileName)
{
  FILE* Hresponse = fopen(ResponseFileName, "rt"); /* Open response text file */
  if ( !Hresponse )
  {
    fprintf(RTTB_DEV_ERROR, "ERROR: Cannot open response file \'%s\'\n", ResponseFileName);
    return;
  }
  while ( !feof(Hresponse) )
  {
    char lineBuff[1024]; /* Maximum size we will allow for a line and therefore a filename */
    size_t length; /* Length of the line */

    /* Read in a line (a filename) */
    if ( !fgets(lineBuff, sizeof(lineBuff)/sizeof(lineBuff[0]), Hresponse) ) break;
    length = strlen(lineBuff);
    length = ChopNewline(lineBuff, length); /* Chop the newline char if there is one */
    /* Add this source filename to our list of files to add */
    if (length)
    {
      unsigned short adderror;

      char* fnCopy = (char*)ddrpmalloc((length+1)*sizeof(char));
      if (fnCopy)
      {
        strcpy(fnCopy, lineBuff);
        rttbutil_RTrimWithLen(fnCopy, length);
        adderror = AddSource(fnCopy, TRUE); /* TRUE = dynamic */
        if ( adderror )
        {
          /* Error adding sourcefile to the list to parse */
          PrintError(adderror, fnCopy);
          ddrpfree(fnCopy);
        }
      }
      else /* fnCopy not allocated */
      {
        PrintError(ERROR_DDRPMALLOC, lineBuff);
      }
    }
  }
  fclose(Hresponse);
}
/*--------------------------------------------------------------------------*/
static size_t ChopNewline(char* LineStr, size_t Length)
{
  size_t i;

  for ( i = 0UL; i < Length; i++ )
  {
    if ( LineStr[i] == '\n' ) /* Got the newline! */
    {
      LineStr[i] = '\0'; /* Make NULL */
      Length--;
      break; /* Do not bother going further */
    }
  }
  return Length;
}
/*--------------------------------------------------------------------------*/
#ifndef NDEBUG /* Debug */
size_t _debug_fwrite(const void* ptr, size_t size, size_t n, FILE* stream)
{
  size_t ret = fwrite(ptr, size, n, stream); /* Call proper function and save return */
  fflush(stream);
  return ret; /* Give result of fwrite() to caller */
}
#endif /*!NDEBUG*/
/*--------------------------------------------------------------------------*/
