/*
Daybo Logic Run-Time TarBall Library And Tools
Copyright (c) 2000-2006, David Duncan Ross Palmer, Daybo Logic
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
      
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
      
    * Neither the name of the Daybo Logic nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

/*
  Most of this code is from the Daybo Logic utilities library,
  however I didn't want RTTB to be dependant on dlutils, some of
  this functionallity could be used from the RTTB library,
  otherwise, it's here.
*/

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif /*HAVE_CONFIG_H*/

#include <stdio.h>

#ifdef HAVE_STDLIB_H
# include <stdlib.h>
#endif /*HAVE_STDLIB_H*/

#ifdef HAVE_STRING_H
# include <string.h>
#endif /*HAVE_STRING_H*/

#ifdef HAVE_LIMITS_H
# include <limits.h>
#endif /*HAVE_LIMITS_H*/

#ifdef RTTB_HDRSTOP
# pragma hdrstop
#endif /*RTTB_HDRSTOP*/

#include "ddrptype.h" /* Types used by Overlord DDRP in the game */
#include "simplemm.h" /* Memory manager */
#include "rttbutil.h"
/*-------------------------------------------------------------------------*/
int rttbutil_GetFileLength(FILE* HFile, unsigned long* PLength)
{
  long int length;
  long int oldFilePos;

  if ( !HFile ) return 0; /*FALSE*/
  oldFilePos = ftell(HFile); /* Save current file offset */
  rewind(HFile); /* Reset any EOF conditions and the like */
  fseek(HFile, 0, SEEK_END); /* Seek to end */
  length = ftell(HFile);
  /* Restore pos */
  if (length == -1L) rewind(HFile);
  fseek(HFile, oldFilePos, SEEK_SET);

  if ( PLength )
    *PLength = (long int)length;

  if ( length == -1L )
    return 0; /*FALSE*/

  return 1; /* TRUE (Success) */
}
/*-------------------------------------------------------------------------*/
int rttbutil_CalcFileBlockSize(unsigned long FileLength, FILE_BLOCK_SIZE* CallerFileBlockSizePtr, const size_t BufferSize)
{
  /* Mathematics by Nigel Crouch, written in C by D. D. R. Palmer. */
  /* initialize structure */
  CallerFileBlockSizePtr->BufferSize = 0UL;
  CallerFileBlockSizePtr->Loops = 0UL;
  CallerFileBlockSizePtr->Remainder = 0UL;

  if (FileLength == 0)
  {
    /* zero file length, handle this error */
    return 0; /*FALSE*/
  }

  if (FileLength <= BufferSize)
    CallerFileBlockSizePtr->BufferSize = FileLength;
  else
    CallerFileBlockSizePtr->BufferSize = BufferSize;

  CallerFileBlockSizePtr->Loops = FileLength / (CallerFileBlockSizePtr->BufferSize);
  CallerFileBlockSizePtr->Remainder = (FileLength - (CallerFileBlockSizePtr->BufferSize) * CallerFileBlockSizePtr->Loops);
  return 1; /*TRUE*/
}
/*-------------------------------------------------------------------------*/
DWORD rttbutil_CalcFileChecksum32(FILE* HFile)
{
  /* Var declarations */
  ULONG lenChkFile, iLoops;
  DWORD chk32 = (DWORD)0UL;
  FILE_BLOCK_SIZE fileBlockSize;
  BYTE *currentData, *remainingData;
  unsigned dataIndex;
  long int stackedFilePos;

  if (!HFile) return chk32;
  stackedFilePos = ftell(HFile);
  if (stackedFilePos == -1L)
    return chk32; /*0UL*/
  if ( !rttbutil_GetFileLength(HFile, &lenChkFile) ) /* Save the length of this file */
    return chk32;

  if (!lenChkFile)
  {
    fclose(HFile);
    return chk32; /* Don't waste time, a zero length file's checksum is always 0 */
  }

  rttbutil_CalcFileBlockSize(lenChkFile, &fileBlockSize, (size_t)4096U); /* Call Nigel's well-magic wonderful fast function!, buffer 4K for read loops */
  currentData = (BYTE*)ddrpmalloc(fileBlockSize.BufferSize); /* Create a buffer for file I/O */

  fseek(HFile, 0, SEEK_SET); /* Reset to file begining */
  for (iLoops=0;iLoops<fileBlockSize.Loops;iLoops++)
  {
    if (fread(currentData, fileBlockSize.BufferSize, 1, HFile) < 1) /* Read data into buffer */
    {
      /* EOF should never occur - AND HAS! */
      fprintf(RTTB_DEV_ERROR, "Unexpected EOF during file checksummer\n");
      abort();
    }
    for ( dataIndex = 0U; dataIndex < fileBlockSize.BufferSize; dataIndex++ ) /* Go through all of the CurrentData read from the file*/
      chk32 += currentData[dataIndex]; /* Add the current byte to the total 32-bit CRC checksum */
  }
  /* Now add the remainder */
  if (fileBlockSize.Remainder)
  {
    remainingData = (BYTE*)ddrpmalloc(fileBlockSize.Remainder); /* Create buffer for remainder */
    fread(remainingData, fileBlockSize.Remainder, 1, HFile); /* Put remaining data in this buffer */
  for ( dataIndex = 0U; dataIndex < fileBlockSize.Remainder; dataIndex++ ) /* go through all of the remaining data byte by byte */
    chk32 += remainingData[dataIndex]; /* Add the remainder to the total */

    ddrpfree(remainingData); /* Deallocate buffer */
  }

  /* Cleanup */
  if (currentData) ddrpfree(currentData);

  rewind(HFile);
  fseek(HFile, stackedFilePos, SEEK_SET); /* Restore file pos */
  return chk32; /* Give the 32-bit CRC value to the caller */
}
/*----------------------------------------------------------------------------*/
char* rttbutil_RTrimWithLen(char* str, size_t len)
{
  /* Trim all right-hand whitespace */

  unsigned int i;

  if ( !str ) return NULL;
  if ( len ) len--; /* Skip NULL terminator but don't wrap to FFFF! */

  for ( i = len; i; i-- ) {
    if ( str[i] == ' ' )
      str[i] = '\0';
    else
      break;
  }
  return str;
}
/*----------------------------------------------------------------------------*/
