/*
Daybo Logic Run-Time TarBall Library And Tools
Copyright (c) 2000-2006, David Duncan Ross Palmer, Daybo Logic
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
      
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
      
    * Neither the name of the Daybo Logic nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

/*
  Small utilliy to verify files in an RTTB
  Author: David Duncan Ross Palmer
  Maintainer: Same
  Date: 6th October 2000 - 27th February 2006
  Email: http://www.daybologic.co.uk/mailddrp/
  Web site for more information: http://www.daybologic.co.uk
*/

#define RTTB_SOURCE /* So we can steal the library's free list */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif /*HAVE_CONFIG_H*/

#include <stdio.h>
#include <assert.h>

#ifdef HAVE_STDDEF_H
# include <stddef.h>
#endif /*HAVE_STDDEF_H*/

#ifdef HAVE_LIMITS_H
# include <limits.h>
#endif /*HAVE_LIMITS_H*/

#ifdef HAVE_STDLIB_H
# include <stdlib.h>
#endif /*HAVE_STDLIB_H*/

#ifdef RTTB_HDRSTOP
# pragma hdrstop
#endif /*RTTB_HDRSTOP*/

#include "ddrptype.h"
#include "simplemm.h" /* The simple memory manager I like to use for rttb */
#include "rttb.h" /* For librttb use */
#include "rttb_ll.h" /* Steal librttb's linked list to save code */
#include "rttb_ver.h" /* Steal year information */

/* Macros and stuff I can change with no hassle */
#define VERSION_MAJOR (0)
#define VERSION_MINOR (1)
#define VERSION_BUILD (0)

/* Files are big endian, swap for little endian */
#ifdef __L_ENDIAN__
#  define MAKEUSABLEDWORD(dw) dw = rttb_Swap32((dw)); /* Swap and replace in code */
#else /* Big endian (native in file format) */
#  define MAKEUSABLEDWORD(dw) /* Remove all byte swapping */
#endif /*__L_ENDIAN__*/
/*-------------------------------------------------------------------------*/
static void Banner(void);
static void Syntax(const char* AppExecName);
static BOOL ProcessArguments(const int argc);
static BOOL Verify(FILE* HRTTB); /* Returns # errors */
static void ReadErrorMsg(void);
static BOOL RttbRead(struct link_ptrs* TargetList, const DWORD N, FILE* StreamPtr);
static BOOL RttbReadFilenames(struct link_ptrs* Filenames, struct link_ptrs* FilenameLengths, const DWORD N, FILE* HRTTB);
static void RttbCleanupFilenames(struct link_ptrs* Filenames);
#ifndef NDEBUG /*Debug*/
static void TrapSync(struct link_ptrs* ll0, struct link_ptrs* ll1);
# define TRAPSYNC(ll0, ll1) TrapSync((ll0), (ll1));
#else /* Not debug */
# define TRAPSYNC(ll0, ll1) /* Strip */
#endif /*!NDEBUG*/
#ifdef __L_ENDIAN__ /* Little endian code only */
void LLSwap(struct link_ptrs* List); /* Byte swaps all linked list elements, provided they are 2 or 4 bytes in size */
#endif /*__L_ENDIAN__*/
/*-------------------------------------------------------------------------*/
int
#ifdef __BCB__
debugmain
#else
main
#endif /*__BCB__*/
(const int argc, const char* argv[])
{
  FILE* HRTTB;

  Banner();
  if ( !ProcessArguments(argc) )
  {
    Syntax(argv[0]);
    return EXIT_FAILURE;
  }

  HRTTB = fopen(argv[1], "rb"); /* Attempt to open rttb */
  if ( !HRTTB )
  {
    printf("ERROR: Failed to open \'%s\'\n", argv[1]);
    return EXIT_FAILURE;
  }
  if ( !Verify(HRTTB) )
    printf("WARNING: Errors were discovered in the archive!\n");

  fclose(HRTTB);
  return EXIT_SUCCESS;
}
/*-------------------------------------------------------------------------*/
static void Banner()
{
  printf("Runtime Tarball verify utility, V%u.%u.%u\n", VERSION_MAJOR, VERSION_MINOR, VERSION_BUILD);
  printf("for RTTB revision %u, built %s %s\n", RTTB_REVISION, __DATE__, __TIME__);
  printf("(C)Copyright 2000");
#if ( RTTB_THISYEAR > 2000 )
  printf(" - %u", RTTB_THISYEAR);
#endif
  printf(", Daybo Logic, all rights reserved.\n\n");
}
/*-------------------------------------------------------------------------*/
static void Syntax(const char* AppExecName)
{
  printf(
    "%s <rttbfile>\n\n",
    (AppExecName) ? (AppExecName) : ("(null)")
  );
  printf("Where rttbfile is the filename of an rttb file created with makerttb\n");
  return;
}
/*-------------------------------------------------------------------------*/
static BOOL ProcessArguments(const int argc)
{
  if ( argc > 2 )
  {
    printf("Too many parameters\n");
    return FALSE;
  }
  else if ( argc < 2 )
  {
    printf("Required parameter missing\n");
    return FALSE;
  }
  return TRUE;
}
/*-------------------------------------------------------------------------*/
static unsigned int Verify(FILE* HRTTB)
{
  /* We're at the top, the file is open and normal */
  BYTE b;
  DWORD dw;
  long int tempFilePos;
  DWORD N;
  BOOL status = TRUE; /* Success/failure of this function */
  /* These linked lists hold the data from the header */
  struct link_ptrs filelengths;
  struct link_ptrs checksums;
  struct link_ptrs fnlens;
  struct link_ptrs filenames;

  /* Init the linked lists */
  rttb_link_ptrs_init(&filelengths, sizeof(unsigned long));
  rttb_link_ptrs_init(&checksums, sizeof(DWORD));
  rttb_link_ptrs_init(&fnlens, sizeof(WORD));
  rttb_link_ptrs_init(&filenames, sizeof(char*));

  printf("Skipping %u bytes of copyright information...\n", RTTB_COPYRIGHTLENGTH);
  fseek(HRTTB, RTTB_COPYRIGHTLENGTH, SEEK_SET); /* Skip copyright */
  if ( fread(&b, sizeof(b), 1, HRTTB) < 1 ) /* Read in revision byte */
  {
    ReadErrorMsg();
    status = FALSE;
    goto giveStatus;
  }
  printf("Revision: %u\n", b);
  if ( b > RTTB_REVISION )
  {
    fprintf(RTTB_DEV_ERROR, "ERROR: This verify utillity can only read revision %u files\n", RTTB_REVISION);
    status = FALSE;
    goto giveStatus;
  }
  if ( fread(&dw, sizeof(dw), 1, HRTTB) < 1 ) /* Read in N */
  {
    ReadErrorMsg();
    status = FALSE;
    goto giveStatus;
  }
  MAKEUSABLEDWORD(dw)
  printf("Number of files in RTTB (according to header): %lu\n", (unsigned long)dw);
  N = dw;
  if ( fread(&dw, sizeof(dw), 1, HRTTB) < 1 ) /* Read fast data marker */
  {
    ReadErrorMsg();
    status = FALSE;
    goto giveStatus;
  }
  MAKEUSABLEDWORD(dw)
  printf("Fast data marker offset: hex %lX (%lu), this will be used.\n", dw, dw);
  tempFilePos = ftell(HRTTB);
  if ( fseek(HRTTB, dw, SEEK_SET) != 0 ) /* Use fast data marker */
  {
    /* Seek failed */
    fprintf(RTTB_DEV_ERROR, "ERROR: Seek error, possibly bad fast data marker!\n");
    return FALSE;
  }
  /* I've verified this fast data pointer actually works, now let's read in
  the table of data */
  printf("Reading table of file lengths...");
  fseek(HRTTB, tempFilePos, SEEK_SET); /* Go back to start of table */
  if ( !RttbRead(&filelengths, N, HRTTB) ) {
    printf("FAIL\n");
    fprintf(RTTB_DEV_ERROR, "ERROR: File length table read incomplete\n");
    status = FALSE;
    goto giveStatus;
  }
  printf("OK\nReading table of checksums...");
  if ( !RttbRead(&checksums, N, HRTTB) ) {
    printf("FAIL\n");
    fprintf(RTTB_DEV_ERROR, "ERROR: Checksum table read incomplete\n");
    status = FALSE;
    goto giveStatus;
  }
  TRAPSYNC(&filelengths, &checksums)
  printf("OK\nReading table of filename lenghs...");
  if ( !RttbRead(&fnlens, N, HRTTB) ) {
    printf("FAIL\n");
    fprintf(RTTB_DEV_ERROR, "ERROR: File name lengths table read incomplete\n");
    status = FALSE;
    goto giveStatus;
  }

#ifdef __L_ENDIAN__
  /* Do byte swaps if neccesary */
  LLSwap(&filelengths);
  LLSwap(&checksums);
  LLSwap(&fnlens);
#endif /*__L_ENDIAN__*/

  TRAPSYNC(&checksums, &fnlens)
  printf("OK\nReading table of filenames...");
  if ( !RttbReadFilenames(&filenames, &fnlens, N, HRTTB) ) {
    printf("FAIL\n");
    fprintf(RTTB_DEV_ERROR, "ERROR: Filenames table read incomplete\n");
    status = FALSE;
    goto giveStatus;
  }
  TRAPSYNC(&fnlens, &filenames)
  printf("OK\n");

giveStatus:
  filelengths.DelAll(&filelengths);
  checksums.DelAll(&checksums);
  fnlens.DelAll(&fnlens);
  RttbCleanupFilenames(&filenames);
  printf("(I can't be arsed with the checksum verifications right now)\n");
  return status;
}
/*-------------------------------------------------------------------------*/
static void ReadErrorMsg()
{
  fprintf(RTTB_DEV_ERROR, "Read error!\n");
}
/*-------------------------------------------------------------------------*/
static BOOL RttbRead(struct link_ptrs* TargetList, const DWORD N, FILE* StreamPtr)
{
  DWORD i;
  size_t nodeSize;
  void* data;
  BOOL ok = TRUE;

  nodeSize = TargetList->NodeSize(TargetList);
  data = ddrpmalloc(nodeSize);
  if ( !data ) {
    fprintf(RTTB_DEV_ERROR, "Out of memory in simple node, temp allocation!  You\'re in serious crap now!\n");
    return FALSE;
  }
  for ( i = (DWORD)0U; i < N; i++ ) {
    if ( fread(data, nodeSize, 1, StreamPtr) < 1 ) {
      ok = FALSE;
      fprintf(RTTB_DEV_ERROR, "File #%lu table read error\n", N);
      break;
    }
    /* Read value, now write to linked list */
    if ( !TargetList->Add(TargetList, data) ) {
      ok = FALSE;
      fprintf(RTTB_DEV_ERROR, "File #%lu table linked list error (out of memory)\n", N);
      break;
    }
  }
  ddrpfree(data);
  return ok;
}
/*-------------------------------------------------------------------------*/
static BOOL RttbReadFilenames(struct link_ptrs* Filenames, struct link_ptrs* FilenameLengths, const DWORD N, FILE* HRTTB)
{
  DWORD i;
  /* We assume validity of N for speed */

  for ( i = (DWORD)0U; i < N; i++ ) {
    WORD fnlen;

    FilenameLengths->Get(FilenameLengths, i, &fnlen);
    if ( fnlen ) {
      char* tmpstr = (char*)ddrpmalloc( fnlen+1 );
      if ( !tmpstr ) {
        printf("ERROR: Out of memory, filename maybe recorded too long in header\n");
        return FALSE;
      }
      fread(tmpstr, fnlen+1, 1, HRTTB);
      tmpstr[fnlen] = '\0'; /* Null terminate */
      /* Now add to linked list of filenames */
      if ( !Filenames->Add(Filenames, &tmpstr) ) {
        ddrpfree(tmpstr);
        fprintf(RTTB_DEV_ERROR, "ERROR: Out of memory adding filename to linked list, filenames too long?\n");
        return FALSE;
      }
    }
  }
  return TRUE;
}
/*-------------------------------------------------------------------------*/
static void RttbCleanupFilenames(struct link_ptrs* Filenames)
{
  while ( Filenames->Count(Filenames) ) {
    char* strptr;

    Filenames->Get(Filenames, 0U, &strptr);
    if ( strptr ) ddrpfree(strptr);
    Filenames->DelByPos(Filenames, 0U);
  }
}
/*-------------------------------------------------------------------------*/
#ifndef NDEBUG /*Debug*/
static void TrapSync(struct link_ptrs* ll0, struct link_ptrs* ll1)
{
  unsigned int c[2];
  assert(ll0); assert(ll1);

  c[0] = ll0->Count(ll0);
  c[1] = ll1->Count(ll1);
  if ( c[0] != c[1] ) {
    fprintf(RTTB_DEV_ERROR, "DEBUG ERROR!  Sync mismatch in tables, should be impossible!!!\n"
                    "c[0]=%u, c[1]=%u\n", c[0], c[1]
    );
  }
}
#endif /*!NDEBUG*/
/*-------------------------------------------------------------------------*/
#ifdef __L_ENDIAN__
void LLSwap(struct link_ptrs* List)
{
  if ( List ) {
    unsigned int i, c;
    size_t elemSize = List->NodeSize(List);
    assert(elemSize == 2 || elemSize == 4); /* Anything else might be dangerous! */
    c = List->Count(List);
    for ( i = 0U; i < c; i++ ) {
      DWORD value;
      /* We know the byte order so it's safe to do these strange type of reads */
      List->Get(List, i, &value);
      if ( elemSize == 2 ) {
        if ( value ) /* Swapping zero is stupid */
          value = rttb_Swap16((WORD)value); /* Do WORD swap */
      }
      else {
        if ( value )
          value = rttb_Swap32(value); /* Do DWORD swap */
      }
      List->Set(List, i, &value);
    }
  }
}
#endif /*__L_ENDIAN__*/
/*-------------------------------------------------------------------------*/
